#ifndef AST_HPP
#define AST_HPP

#include <iostream>
#include <list>
#include <memory>
#include <variant>
#include <vector>

namespace ast {
enum class CompUnitKind {
    DECL,
    FUNC_DEF,
};

enum class DeclKind {
    CONST,
    VAR,
};

enum class InitValKind {
    VAL,
    ARRAY,
};

enum class BlockItemKind {
    DECL,
    STMT,
};

enum class StmtKind {
    ASSIGN,
    EXP,
    BLOCK,
    IF,
    WHILE,
    BREAK,
    CONTINUE,
    RETURN,
    PUTINT,
    PUTCH,
    PUTFLOAT,
    PUTARRAY,
    PUTFARRAY,
    PUTF,
    STARTTIME,
    STOPTIME,
};

enum class ExpKind {
    BINOP,
    COND,
    ID,
    INTCONST,
    FLOATCONST,
    NOT,
    MINUS,
    CALL,
    ARRAY,
    GETINT,
    GETCH,
    GETFLOAT,
    GETARRAY,
    GETFARRAY,
};

enum class BaseType {
    INT,
    FLOAT,
};

enum class FuncType {
    VOID,
    INT,
    FLOAT,
};

enum class BinOp {
    ADD,
    SUB,
    MUL,
    DIV,
    MOD,
};

enum class RelOp {
    GT,
    LT,
    GE,
    LE,
    EQ,
    NE,
    AND,
    OR,
};

class Pos;
class Prog;
class CompUnitList;
class CompUnit;

class Decl;

class ConstDecl;
class ConstDefList;
class ConstDef;
class VarDecl;
class VarDefList;
class VarDef;
class InitVal;
class InitValList;

class FuncDef;
class FuncFParamList;
class FuncFParam;

class Block;
class BlockItemList;
class BlockItem;
class Stmt;

class ExpList;
class Exp;

class Pos {
public:
    int line;
    int pos;
    Pos(int line, int pos) : line(line), pos(pos) {}
};

class Prog {
public:
    CompUnitList *comp_unit_list;
    Prog(CompUnitList *comp_unit_list) : comp_unit_list(comp_unit_list) {}
};

class CompUnitList {
public:
    // std::vector<CompUnit *> comp_units;
    std::list<CompUnit *> comp_units;
    CompUnitList() = default;
    void push_back(CompUnit *comp_unit) { comp_units.push_back(comp_unit); }
    void push_front(CompUnit *comp_unit) { comp_units.push_front(comp_unit); }
};

class CompUnit {
public:
    Pos *pos;
    CompUnitKind kind;
    union {
        Decl *decl;
        FuncDef *func_def;
    } u;

    CompUnit() = default;
};

class Decl {
public:
    Pos *pos;
    DeclKind kind;
    union {
        ConstDecl *const_decl;
        VarDecl *var_decl;
    } u;

    Decl() = default;
};

class ConstDecl {
public:
    Pos *pos;
    BaseType type;
    ConstDefList *const_defs;

    ConstDecl() = default;
};

class ConstDefList {
public:
    std::list<ConstDef *> const_defs;
    ConstDefList() = default;
    void push_back(ConstDef *const_def) { const_defs.push_back(const_def); }
    void push_front(ConstDef *const_def) { const_defs.push_front(const_def); }
};

class ConstDef {
public:
    Pos *pos;
    std::string *id;
    ExpList *array_dim;
    InitVal *init_val;

    ConstDef() = default;
};

class InitVal {
public:
    Pos *pos;
    InitValKind kind;
    union {
        Exp *val;
        InitValList *array;
    } u;
};

class InitValList {
public:
    std::vector<InitVal *> array;
    InitValList() = default;

    void push_back(InitVal *init_val) { array.push_back(init_val); }
    void push_front(InitVal *init_val) { array.insert(array.begin(), init_val); }
};

class VarDecl {
public:
    Pos *pos;
    BaseType type;
    VarDefList *var_defs;

    VarDecl() = default;
};

class VarDefList {
public:
    std::list<VarDef *> var_defs;
    VarDefList() = default;
    void push_back(VarDef *var_def) { var_defs.push_back(var_def); }
    void push_front(VarDef *var_def) { var_defs.push_front(var_def); }
};

class VarDef {
public:
    Pos *pos;
    std::string *id;
    ExpList *array_dim;
    InitVal *init_val;

    VarDef() = default;
};

class FuncDef {
public:
    Pos *pos;
    FuncType ret_type;
    std::string *id;
    FuncFParamList *f_param_list;
    Block *block;

    FuncDef() = default;
};

class FuncFParamList {
public:
    std::list<FuncFParam *> f_params;
    FuncFParamList() = default;
    void push_back(FuncFParam *f_param) { f_params.push_back(f_param); }
    void push_front(FuncFParam *f_param) { f_params.push_front(f_param); }
};

class FuncFParam {
public:
    Pos *pos;
    BaseType type;
    std::string *id;
    bool is_array;
    ExpList *array_dim;

    FuncFParam() = default;
};

class Block {
public:
    Pos *pos;
    BlockItemList *block_item_list;

    Block() = default;
};

class BlockItemList {
public:
    std::list<BlockItem *> block_items;
    BlockItemList() = default;
    void push_back(BlockItem *block_item) { block_items.push_back(block_item); }
    void push_front(BlockItem *block_item) { block_items.push_front(block_item); }
};

class BlockItem {
public:
    Pos *pos;
    BlockItemKind kind;
    union {
        Decl *decl;
        Stmt *stmt;
    } u;

    BlockItem() = default;
};

class Stmt {
public:
    Pos *pos;
    StmtKind kind;
    union {
        struct {
            Exp *lhs;
            Exp *rhs;
        } assign;
        Exp *exp;
        Block *block;
        struct {
            Exp *cond;
            Stmt *thenStmt;
            Stmt *elseStmt;
        } if_stmt;
        struct {
            Exp *cond;
            Stmt *body;
        } while_stmt;
        struct {
            Exp *array;
            Exp *out_len;
        } putarray;
        struct {
            std::string *fmt;
            ExpList *args;
        } putf;
    } u;
};

class ExpList {
public:
    std::list<Exp *> exps;
    ExpList() = default;
    void push_back(Exp *exp) { exps.push_back(exp); }
    void push_front(Exp *exp) { exps.push_front(exp); }
};

class Exp {
public:
    Pos *pos;
    ExpKind kind;
    union {
        struct {
            BinOp op;
            Exp *lhs;
            Exp *rhs;
        } binop;
        struct {
            RelOp op;
            Exp *lhs;
            Exp *rhs;
        } cond;
        std::string *id;
        int int_val;
        std::string *float_val;
        Exp *exp;
        struct {
            std::string *funcname;
            ExpList *args;
        } call;
        struct {
            Exp *array;
            ExpList *index;
        } array;
    } u;
};

Pos *make_pos(int line, int pos);
Prog *make_prog(CompUnitList *comp_unit_list);
CompUnitList *make_comp_unit_list(CompUnit *comp_unit, CompUnitList *comp_unit_list);
CompUnit *make_unit_decl(Pos *pos, Decl *decl);
CompUnit *make_unit_func(Pos *pos, FuncDef *func_def);
Decl *make_decl_const(Pos *pos, ConstDecl *const_decl);
Decl *make_decl_var(Pos *pos, VarDecl *var_decl);
ConstDecl *make_const_decl(Pos *pos, BaseType type, ConstDefList *const_defs);
ConstDefList *make_const_def_list(ConstDef *const_def, ConstDefList *const_defs);
ConstDef *make_const_def(Pos *pos, std::string *id, ExpList *array_dim, InitVal *init_val);
InitVal *make_init_val_exp(Pos *pos, Exp *val);
InitVal *make_init_val_array(Pos *pos, InitValList *array);
InitValList *make_init_val_list(InitVal *init_val, InitValList *init_val_list);
VarDecl *make_var_decl(Pos *pos, BaseType type, VarDefList *var_defs);
VarDefList *make_var_def_list(VarDef *var_def, VarDefList *var_defs);
VarDef *make_var_def(Pos *pos, std::string *id, ExpList *array_dim, InitVal *init_val);
FuncDef *make_func_def(Pos *pos, FuncType ret_type, std::string *id, FuncFParamList *f_param_list, Block *block);
FuncFParamList *make_func_f_param_list(FuncFParam *f_param, FuncFParamList *f_params);
FuncFParam *make_func_f_param(Pos *pos, BaseType type, std::string *id, bool is_array, ExpList *array_dim);
Block *make_block(Pos *pos, BlockItemList *block_item_list);
BlockItemList *make_block_item_list(BlockItem *block_item, BlockItemList *block_item_list);
BlockItem *make_block_item_decl(Pos *pos, Decl *decl);
BlockItem *make_block_item_stmt(Pos *pos, Stmt *stmt);

Stmt *make_assign_stmt(Pos *pos, Exp *lhs, Exp *rhs);
Stmt *make_exp_stmt(Pos *pos, Exp *exp);
Stmt *make_block_stmt(Pos *pos, Block *block);
Stmt *make_if_stmt(Pos *pos, Exp *cond, Stmt *thenStmt, Stmt *elseStmt);
Stmt *make_while_stmt(Pos *pos, Exp *cond, Stmt *body);
Stmt *make_break_stmt(Pos *pos);
Stmt *make_continue_stmt(Pos *pos);
Stmt *make_return_stmt(Pos *pos, Exp *exp);
Stmt *make_putint_stmt(Pos *pos, Exp *exp);
Stmt *make_putch_stmt(Pos *pos, Exp *exp);
Stmt *make_putfloat_stmt(Pos *pos, Exp *exp);
Stmt *make_putarray_stmt(Pos *pos, Exp *array, Exp *index);
Stmt *make_putfarray_stmt(Pos *pos, Exp *array, Exp *index);
Stmt *make_putf_stmt(Pos *pos, std::string *fmtPtr, ExpList *args);
Stmt *make_starttime_stmt(Pos *pos);
Stmt *make_stoptime_stmt(Pos *pos);

ExpList *make_exp_list(Exp *exp, ExpList *exp_list);
Exp *make_binop_exp(Pos *pos, BinOp op, Exp *lhs, Exp *rhs);
Exp *make_cond_exp(Pos *pos, RelOp op, Exp *lhs, Exp *rhs);
Exp *make_id_exp(Pos *pos, std::string *id);
Exp *make_intconst_exp(Pos *pos, int int_val);
Exp *make_floatconst_exp(Pos *pos, std::string *float_val);
Exp *make_not_exp(Pos *pos, Exp *exp);
Exp *make_minus_exp(Pos *pos, Exp *exp);
Exp *make_call_exp(Pos *pos, std::string *funcname, ExpList *args);
Exp *make_array_exp(Pos *pos, Exp *array, ExpList *index);
Exp *make_getint_exp(Pos *pos);
Exp *make_getch_exp(Pos *pos);
Exp *make_getfloat_exp(Pos *pos);
Exp *make_getarray_exp(Pos *pos, Exp *exp);
Exp *make_getfarray_exp(Pos *pos, Exp *exp);

} // namespace ast

#endif // AST_HPP