#ifndef TREEP_HPP
#define TREEP_HPP

#include "env.hpp"
#include "table.hpp"
#include "temp.hpp"
#include <iostream>
#include <vector>

namespace irp {
enum class GlobalDefKind {
    CONST_ARR,
    VAR,
    ARR,
};

enum class StmtKind {
    SEQ,
    LABEL,
    MOVE,
    EXP,
    RETURN,
    JUMP,
    CJUMP,
    VOID_CALL,
    PUTINT,
    PUTCH,
    PUTFLOAT,
    PUTARRAY,
    PUTFARRAY,
    PUTF,
    STARTTIME,
    STOPTIME,
    MEMSET,
};

enum class ExpKind {
    BINOP,
    MEM,
    TEMP,
    ESEQ,
    NAME,
    CONSTI,
    CONSTF,
    CALL,
    CAST,
    GETINT,
    GETCH,
    GETFLOAT,
    GETARRAY,
    GETFARRAY,
    ALLOCA,
};

enum class ExpType {
    INT,
    FLOAT,
};

enum class FuncType {
    VOID,
    INT,
    FLOAT,
};

enum class BinOp {
    ADD,
    SUB,
    MUL,
    DIV,
    MOD,
};

enum class RelOp {
    GT,
    LT,
    GE,
    LE,
    EQ,
    NE,
};

class Prog;
class GlobalDef;
class LocalConstArrayDef;
class FuncDef;
class Stmt;
class Exp;
class StmtList;

using FuncDefList = std::vector<FuncDef *>;
using ExpList = std::vector<Exp *>;

class Prog {
public:
    FuncDefList *funcs;
    std::vector<GlobalDef *> globalDefs;
    std::vector<LocalConstArrayDef *> localConstArrayDefs;

    Prog() = default;
};

class GlobalDef {
public:
    GlobalDefKind kind;
    temp::Label *label;
    std::string *llvmStr;
    std::string *armStr;
};

class LocalConstArrayDef {
public:
    temp::Label *label;
    std::string *llvmStr;
    std::string *armStr;
};

class FuncDef {
public:
    std::string *name;
    temp::TempList *params;
    Stmt *body;
    FuncType type;
};

class Stmt {
public:
    StmtKind kind;
    union {
        struct {
            Stmt *left;
            Stmt *right;
        } seq;
        temp::Label *label;
        struct {
            Exp *dst;
            Exp *src;
        } move;
        Exp *exp;
        struct {
            RelOp op;
            Exp *left, *right;
            temp::Label *t;
            temp::Label *f;
        } cjump;
        temp::Label *jump;
        struct {
            std::string *name;
            ExpList *args;
        } void_call;
        struct {
            Exp *array;
            Exp *outlen;
        } putarray;
        struct {
            std::string *fmt;
            ExpList *args;
        } putf;
        struct {
            temp::Temp *array;
            int offset;
            int size;
        } memset;
    } u;
};

class Exp {
public:
    ExpKind kind;
    ExpType type;
    union {
        struct {
            BinOp op;
            Exp *left;
            Exp *right;
        } binop;
        Exp *mem;
        temp::Temp *temp;
        struct {
            Stmt *stmt;
            Exp *exp;
        } eseq;
        temp::Label *name;
        int consti;
        float constf;
        struct {
            std::string *name;
            ExpList *args;
        } call;
        Exp *cast;
        Exp *getarray;
        struct {
            int size;
            int frameOffset;
        } alloca;
    } u;
};

class StmtList {
public:
    Stmt *head;
    StmtList *tail;
    StmtList(Stmt *head, StmtList *tail) : head(head), tail(tail) {}
};

FuncDef *getFuncDef(std::string *name, temp::TempList *params, Stmt *body, FuncType type);

GlobalDef *GlobalConstArrDef(temp::Label *label, std::string *llvmStr, std::string *armStr);
GlobalDef *GlobalVarDef(temp::Label *label, std::string *llvmStr, std::string *armStr);
GlobalDef *GlobalArrDef(temp::Label *label, std::string *llvmStr, std::string *armStr);

LocalConstArrayDef *LocalConstArrDef(temp::Label *label, std::string *llvmStr, std::string *armStr);

Stmt *SeqStmt(Stmt *left, Stmt *right);
Stmt *LabelStmt(temp::Label *label);
Stmt *MoveStmt(Exp *dst, Exp *src);
Stmt *ExpStmt(Exp *exp);
Stmt *ReturnStmt(Exp *exp);
Stmt *JumpStmt(temp::Label *jump);
Stmt *CJumpStmt(RelOp op, Exp *left, Exp *right, temp::Label *t, temp::Label *f);
Stmt *VoidCallStmt(std::string *name, ExpList *args);
Stmt *PutintStmt(Exp *exp);
Stmt *PutchStmt(Exp *exp);
Stmt *PutfloatStmt(Exp *exp);
Stmt *PutarrayStmt(Exp *array, Exp *outlen);
Stmt *PutfarrayStmt(Exp *array, Exp *outlen);
Stmt *PutfStmt(std::string *fmt, ExpList *args);
Stmt *StarttimeStmt();
Stmt *StoptimeStmt();
Stmt *MemsetStmt(temp::Temp *array, int offset, int size);

Exp *BinOpExp(BinOp op, Exp *left, Exp *right);
Exp *MemExp(Exp *mem, ExpType type);
Exp *TempExp(temp::Temp *temp);
Exp *EseqExp(Stmt *stmt, Exp *exp);
Exp *NameExp(temp::Label *name);
Exp *ConstIExp(int consti);
Exp *ConstFExp(float constf);
Exp *CallExp(std::string *name, ExpList *args, ExpType type);
Exp *CastExp(Exp *exp, ExpType type);
Exp *GetintExp();
Exp *GetchExp();
Exp *GetfloatExp();
Exp *GetarrayExp(Exp *array);
Exp *GetfarrayExp(Exp *array);
Exp *AllocaExp(int size, int offset);
} // namespace irp

#endif // TREEP_HPP