#ifndef LLVM_HPP
#define LLVM_HPP

#include "temp.hpp"

namespace mir {
enum class BinOp {
    ADD_I,
    ADD_F,
    SUB_I,
    SUB_F,
    MUL_I,
    MUL_F,
    DIV_I,
    DIV_F,
    MOD,
};

enum class RelOp {
    GT_I,
    GT_F,
    LT_I,
    LT_F,
    GE_I,
    GE_F,
    LE_I,
    LE_F,
    EQ_I,
    EQ_F,
    NE_I,
    NE_F,
};

enum class InstKind {
    BINOP,
    MOVE,
    LOAD,
    STORE,
    LABEL,
    JUMP,
    CMP,
    CJUMP,
    CALL,
    RET,
    PHI,
    ALLOCA,
    INT_TO_PTR,
    PTR_TO_INT,
    INT_TO_FLOAT,
    FLOAT_TO_INT,
    GET_ELEMENT_PTR,
};

class PhiOperand {
public:
    temp::Operand *op;
    temp::Label *label;
    PhiOperand(temp::Operand *op, temp::Label *label) : op(op), label(label) {}
};

using PhiOperandList = std::vector<PhiOperand>;

class Instuction {
public:
    virtual ~Instuction() = default;
    virtual InstKind getInstKind() const = 0;
};

class BinOpInst : public Instuction {
public:
    BinOp op;
    temp::Operand *dst, *lhs, *rhs;
    BinOpInst(BinOp op, temp::Operand *dst, temp::Operand *lhs, temp::Operand *rhs)
        : op(op), dst(dst), lhs(lhs), rhs(rhs) {}
    InstKind getInstKind() const override { return InstKind::BINOP; }
};

class MoveInst : public Instuction {
public:
    temp::Operand *dst, *src;
    MoveInst(temp::Operand *dst, temp::Operand *src) : dst(dst), src(src) {}
    InstKind getInstKind() const override { return InstKind::MOVE; }
};

class LoadInst : public Instuction {
public:
    temp::Operand *dst, *srcPtr;
    LoadInst(temp::Operand *dst, temp::Operand *srcPtr) : dst(dst), srcPtr(srcPtr) {}
    InstKind getInstKind() const override { return InstKind::LOAD; }
};

class StoreInst : public Instuction {
public:
    temp::Operand *dstPtr, *src;
    StoreInst(temp::Operand *dstPtr, temp::Operand *src) : dstPtr(dstPtr), src(src) {}
    InstKind getInstKind() const override { return InstKind::STORE; }
};

class LabelInst : public Instuction {
public:
    temp::Label *label;
    LabelInst(temp::Label *label) : label(label) {}
    InstKind getInstKind() const override { return InstKind::LABEL; }
};

class JumpInst : public Instuction {
public:
    temp::Label *label;
    JumpInst(temp::Label *label) : label(label) {}
    InstKind getInstKind() const override { return InstKind::JUMP; }
};

class CmpInst : public Instuction {
public:
    RelOp op;
    temp::Operand *dst, *lhs, *rhs;
    CmpInst(RelOp op, temp::Operand *dst, temp::Operand *lhs, temp::Operand *rhs)
        : op(op), dst(dst), lhs(lhs), rhs(rhs) {}
    InstKind getInstKind() const override { return InstKind::CMP; }
};

class CJumpInst : public Instuction {
public:
    temp::Operand *cond;
    temp::Label *t, *f;
    CJumpInst(temp::Operand *cond, temp::Label *t, temp::Label *f) : cond(cond), t(t), f(f) {}
    InstKind getInstKind() const override { return InstKind::CJUMP; }
};

class CallInst : public Instuction {
public:
    temp::Operand *dst; // nullptr if void
    std::string funcName;
    temp::OperandList *args;
    CallInst(temp::Operand *dst, const std::string &funcName, temp::OperandList *args)
        : dst(dst), funcName(funcName), args(args) {}
    InstKind getInstKind() const override { return InstKind::CALL; }
};

class RetInst : public Instuction {
public:
    temp::Operand *ret;
    RetInst(temp::Operand *ret) : ret(ret) {}
    InstKind getInstKind() const override { return InstKind::RET; }
};

class PhiInst : public Instuction {
public:
    temp::Operand *dst;
    PhiOperandList *labels;
    PhiInst(temp::Operand *dst, PhiOperandList *labels)
        : dst(dst), labels(labels) {}
    InstKind getInstKind() const override { return InstKind::PHI; }
};

class AllocaInst : public Instuction {
public:
    temp::Operand *dst;
    int size;
    AllocaInst(temp::Operand *dst, int size) : dst(dst), size(size) {}
    InstKind getInstKind() const override { return InstKind::ALLOCA; }
};

class IntToPtrInst : public Instuction {
public:
    temp::Operand *dst;
    temp::Operand *src;
    IntToPtrInst(temp::Operand *dst, temp::Operand *src) : dst(dst), src(src) {}
    InstKind getInstKind() const override { return InstKind::INT_TO_PTR; }
};

class PtrToIntInst : public Instuction {
public:
    temp::Operand *dst;
    temp::Operand *src;
    PtrToIntInst(temp::Operand *dst, temp::Operand *src) : dst(dst), src(src) {}
    InstKind getInstKind() const override { return InstKind::PTR_TO_INT; }
};

class IntToFloatInst : public Instuction {
public:
    temp::Operand *dst;
    temp::Operand *src;
    IntToFloatInst(temp::Operand *dst, temp::Operand *src) : dst(dst), src(src) {}
    InstKind getInstKind() const override { return InstKind::INT_TO_FLOAT; }
};

class FloatToIntInst : public Instuction {
public:
    temp::Operand *dst;
    temp::Operand *src;
    FloatToIntInst(temp::Operand *dst, temp::Operand *src) : dst(dst), src(src) {}
    InstKind getInstKind() const override { return InstKind::FLOAT_TO_INT; }
};

class GetElementPtrInst : public Instuction {
public:
    temp::Operand *dst;
    temp::Operand *srcPtr;
    temp::Operand *indices;
    GetElementPtrInst(temp::Operand *dst, temp::Operand *srcPtr, temp::Operand *indices)
        : dst(dst), srcPtr(srcPtr), indices(indices) {}
    InstKind getInstKind() const override { return InstKind::GET_ELEMENT_PTR; }
};

} // namespace mir

#endif // LLVM_HPP