#include "ast2irp.hpp"
#include "env.hpp"
#include "irp.hpp"
#include "temp.hpp"

#include <cassert>
#include <cstddef>
#include <cstdlib>
#include <iostream>
#include <numeric>

#define IRPGEN_DEBUG
// #undef IRPGEN_DEBUG

namespace irpgen {
union uf {
    int i;
    float f;
    unsigned u;

    uf(int i) : i(i) {}
    uf(float f) : f(f) {}
};
unsigned float2unsigned(float f) {
    uf u(f);
    return u.u;
}

class PatchList {
public:
    temp::Label **head;
    PatchList *tail;
    PatchList(temp::Label **head, PatchList *tail) : head(head), tail(tail) {}
};

void doPatch(PatchList *tList, temp::Label *label) {
    for (PatchList *p = tList; p; p = p->tail) {
        *p->head = label;
    }
}

PatchList *joinPatch(PatchList *first, PatchList *second) {
    if (!first) {
        return second;
    }
    PatchList *p = first;
    while (p->tail) {
        p = p->tail;
    }
    p->tail = second;
    return first;
}

class TrCx {
public:
    PatchList *trues;
    PatchList *falses;
    irp::Stmt *stmt;

    TrCx(PatchList *trues, PatchList *falses, irp::Stmt *stmt) : trues(trues), falses(falses), stmt(stmt) {}
};

class TrExp {
public:
    enum { TrEx, TrNx, TrCx } kind;
    union {
        irp::Exp *ex;
        irp::Stmt *nx;
        class TrCx *cx;
    } u;

    TrExp(irp::Exp *ex) : kind(TrEx) { u.ex = ex; }
    TrExp(irp::Stmt *nx) : kind(TrNx) { u.nx = nx; }
    TrExp(PatchList *trues, PatchList *falses, irp::Stmt *stmt) : kind(TrCx) {
        u.cx = new class TrCx(trues, falses, stmt);
    }
};

irp::Exp *unEx(TrExp *exp) {
    if (!exp) {
        return nullptr;
    }
    switch (exp->kind) {
    case TrExp::TrEx:
        return exp->u.ex;
    case TrExp::TrNx:
        return irp::EseqExp(exp->u.nx, irp::ConstIExp(0));
    case TrExp::TrCx: {
        temp::Temp *r = temp::newTemp(temp::TempType::INT);
        temp::Label *t = temp::newLabel();
        temp::Label *f = temp::newLabel();
        temp::Label *e = temp::newLabel();
        doPatch(exp->u.cx->trues, t);
        doPatch(exp->u.cx->falses, f);
        return irp::EseqExp(
            irp::SeqStmt(
                exp->u.cx->stmt,
                irp::SeqStmt(irp::LabelStmt(t),
                             irp::SeqStmt(irp::MoveStmt(irp::TempExp(r), irp::ConstIExp(1)),
                                          irp::SeqStmt(irp::JumpStmt(e),
                                                       irp::SeqStmt(irp::LabelStmt(f),
                                                                    irp::SeqStmt(irp::MoveStmt(irp::TempExp(r),
                                                                                               irp::ConstIExp(0)),
                                                                                 irp::LabelStmt(e))))))),
            irp::TempExp(r));
    }
    default:
        assert(0);
    }
}

irp::Stmt *unNx(TrExp *exp) {
    if (!exp) {
        return nullptr;
    }
    switch (exp->kind) {
    case TrExp::TrEx:
        return irp::ExpStmt(exp->u.ex);
    case TrExp::TrNx:
        return exp->u.nx;
    case TrExp::TrCx:
        return exp->u.cx->stmt;
    default:
        assert(0);
    }
}

TrCx *unCx(TrExp *exp) {
    if (!exp) {
        return nullptr;
    }
    switch (exp->kind) {
    case TrExp::TrEx: {
        irp::Stmt *stmt = irp::CJumpStmt(irp::RelOp::NE, exp->u.ex, irp::ConstIExp(0), nullptr, nullptr);
        PatchList *trues = new PatchList(&(stmt->u.cjump.t), nullptr);
        PatchList *falses = new PatchList(&(stmt->u.cjump.f), nullptr);
        return new class TrCx(trues, falses, stmt);
    }
    case TrExp::TrCx:
        return exp->u.cx;
    case TrExp::TrNx:
    default:
        assert(0);
    }

    return nullptr;
}

TrExp *TrStmtList(TrExp *first, TrExp *second) {
    if (!first && !second) {
        return nullptr;
    }
    if (!first) {
        return new TrExp(unNx(second));
    }
    if (!second) {
        return new TrExp(unNx(first));
    }
    return new TrExp(irp::SeqStmt(unNx(first), unNx(second)));
}

class TrExpList {
public:
    std::list<TrExp *> exps;
    TrExpList() = default;
    void push_back(TrExp *exp) { exps.push_back(exp); }
    void push_front(TrExp *exp) { exps.push_front(exp); }
};

irp::ExpList *unTrExpList(TrExpList *trExpList) {
    if (!trExpList) {
        return nullptr;
    }
    irp::ExpList *expList = new irp::ExpList();
    for (auto trExp : trExpList->exps) {
        irp::Exp *exp = unEx(trExp);
        expList->push_back(exp);
    }
    return expList;
}

class ExpTy {
public:
    env::Type *ty;
    TrExp *exp;
    ExpTy(env::Type *ty, TrExp *exp) : ty(ty), exp(exp) {}
};

class LoopInfo {
public:
    temp::Label *test;
    temp::Label *done;
    LoopInfo(temp::Label *test, temp::Label *done) : test(test), done(done) {}
};

class ConstInitValInfo {
public:
    bool isFilledZero;
    int size;
    std::string *llvmStr;
    std::string *armStr;
    ConstInitValInfo(bool isFilledZero, int size, std::string *llvmStr, std::string *armStr)
        : isFilledZero(isFilledZero), size(size), llvmStr(llvmStr), armStr(armStr) {}
};

class InitValInfo {
public:
    int size;
    TrExp *stmt;
    std::string *llvmStr;
    std::string *armStr;
    InitValInfo(int size, TrExp *stmt, std::string *llvmStr, std::string *armStr)
        : size(size), stmt(stmt), llvmStr(llvmStr), armStr(armStr) {}
};

void AST2IRP::translate() {
#ifdef IRPGEN_DEBUG
    std::cerr << "AST2IRP::translate()" << std::endl;
#endif
    this->result = new irp::Prog();
    this->beforeMainDefs = nullptr;
    auto funcList = transCompUnitList(this->root->comp_unit_list);

    // Add before main definitions
    for (auto func : *funcList) {
        if (*func->name == "main") {
            func->body = unNx(TrStmtList(this->beforeMainDefs, new TrExp(func->body)));
            break;
        }
    }

    this->result->funcs = funcList;
}

irp::FuncDefList *AST2IRP::transCompUnitList(ast::CompUnitList *compUnitList) {
#ifdef IRPGEN_DEBUG
    std::cerr << "AST2IRP::transCompUnitList()" << std::endl;
#endif
    if (!compUnitList) {
        return nullptr;
    }
    irp::FuncDefList *res = new irp::FuncDefList();
    for (auto compUnit : compUnitList->comp_units) {
        switch (compUnit->kind) {
        case ast::CompUnitKind::FUNC_DEF: {
            irp::FuncDef *funcDef = transFuncDef(compUnit->u.func_def);
            if (funcDef) {
                res->push_back(funcDef);
            }
            break;
        }
        case ast::CompUnitKind::DECL: {
            transGlobalDecl(compUnit->u.decl);
            break;
        }
        default:
            assert(0);
        }
    }
    return res;
}

irp::FuncDef *AST2IRP::transFuncDef(ast::FuncDef *funcDef) {
#ifdef IRPGEN_DEBUG
    std::cerr << "AST2IRP::transFuncDef() for " << *funcDef->id << std::endl;
#endif
    if (!funcDef) {
        return nullptr;
    }
    this->curFuncName = funcDef->id;
    this->curFuncStackOffset = 0;
    this->curFuncAllocas = nullptr;

    irp::FuncType retType;
    env::FuncRetType envRetType;
    switch (funcDef->ret_type) {
    case ast::FuncType::VOID: {
        retType = irp::FuncType::VOID;
        envRetType = env::FuncRetType::VOID;
        break;
    }
    case ast::FuncType::INT: {
        retType = irp::FuncType::INT;
        envRetType = env::FuncRetType::INT;
        break;
    }
    case ast::FuncType::FLOAT: {
        retType = irp::FuncType::FLOAT;
        envRetType = env::FuncRetType::FLOAT;
        break;
    }
    default:
        assert(0);
    }

    auto [fParamList, fParamTempList] = transFuncFParamList(funcDef->f_param_list);
    this->funcTable.enter(*funcDef->id, env::FuncEntry(envRetType, funcDef->id, fParamList));
    auto body = transBlock(funcDef->block);

    return irp::getFuncDef(funcDef->id, fParamTempList, unNx(TrStmtList(this->curFuncAllocas, body)), retType);
}

std::pair<std::vector<env::TyKind> *, temp::TempList *> AST2IRP::transFuncFParamList(ast::FuncFParamList *paramList) {
#ifdef IRPGEN_DEBUG
    std::cerr << "AST2IRP::transFuncFParamList()" << std::endl;
#endif
    if (!paramList) {
        return {nullptr, nullptr};
    }
    std::vector<env::TyKind> *res = new std::vector<env::TyKind>();
    temp::TempList *tempList = new temp::TempList();
    for (auto param : paramList->f_params) {
        auto [tyKind, temp] = transFuncFParam(param);
        res->push_back(tyKind);
        tempList->push_back(temp);
    }
    return {res, tempList};
}

std::pair<env::TyKind, temp::Temp *> AST2IRP::transFuncFParam(ast::FuncFParam *param) {
#ifdef IRPGEN_DEBUG
    std::cerr << "AST2IRP::transFuncFParam() for " << *param->id << ", is_array: " << param->is_array << std::endl;
#endif
    if (!param) {
        return {env::TyKind::VOID, nullptr};
    }

    if (param->is_array) {
        TrExpList *dims = transExpList(param->array_dim);
        irp::ExpList *irpDims = unTrExpList(dims);
        std::vector<int> *dimList = getArrayDimList(irpDims);
        if (!dimList) {
            dimList = new std::vector<int>();
            dimList->push_back(env::FPARAM_ARRAY_FIRST_DIM);
        } else {
            dimList->insert(dimList->begin(), env::FPARAM_ARRAY_FIRST_DIM);
        }

        switch (param->type) {
        case ast::BaseType::INT: {
            temp::Temp *temp = temp::newTemp(temp::TempType::INT_PTR);
            this->localTable.enter(*param->id, env::LocalIArrEntry(temp, dimList));
            return {env::TyKind::INT_PTR, temp};
        }
        case ast::BaseType::FLOAT: {
            temp::Temp *temp = temp::newTemp(temp::TempType::FLOAT_PTR);
            this->localTable.enter(*param->id, env::LocalFArrEntry(temp, dimList));
            return {env::TyKind::FLOAT_PTR, temp};
        }
        default:
            assert(0);
        }
    } else {
        switch (param->type) {
        case ast::BaseType::INT: {
            temp::Temp *temp = temp::newTemp(temp::TempType::INT);
            this->localTable.enter(*param->id, env::LocalIntEntry(temp));
            return {env::TyKind::INT, temp};
        }
        case ast::BaseType::FLOAT: {
            temp::Temp *temp = temp::newTemp(temp::TempType::FLOAT);
            this->localTable.enter(*param->id, env::LocalFloatEntry(temp));
            return {env::TyKind::FLOAT, temp};
        }
        default:
            assert(0);
        }
    }
}

std::vector<int> *AST2IRP::getArrayDimList(irp::ExpList *dimList) {
#ifdef IRPGEN_DEBUG
    std::cerr << "AST2IRP::getArrayDimList()" << std::endl;
#endif
    if (!dimList) {
        return nullptr;
    }
    std::vector<int> *res = new std::vector<int>();
    for (auto dim : *dimList) {
        if (dim->kind != irp::ExpKind::CONSTI && dim->kind != irp::ExpKind::CONSTF) {
            assert(0);
        }
        switch (dim->kind) {
        case irp::ExpKind::CONSTI: {
            res->push_back(dim->u.consti);
            break;
        }
        case irp::ExpKind::CONSTF: {
            res->push_back(static_cast<int>(dim->u.constf));
            break;
        }
        default:
            assert(0);
        }
    }
    return res;
}

void AST2IRP::transGlobalDecl(ast::Decl *decl) {
#ifdef IRPGEN_DEBUG
    std::cerr << "AST2IRP::transGlobalDecl()" << std::endl;
#endif
    if (!decl) {
        return;
    }
    switch (decl->kind) {
    case ast::DeclKind::CONST: {
        transGlobalConstDecl(decl->u.const_decl);
        break;
    }
    case ast::DeclKind::VAR: {
        this->beforeMainDefs = TrStmtList(this->beforeMainDefs, transGlobalVarDecl(decl->u.var_decl));
        break;
    }
    default:
        assert(0);
    }
}

void AST2IRP::transGlobalConstDecl(ast::ConstDecl *constDecl) {
#ifdef IRPGEN_DEBUG
    std::cerr << "AST2IRP::transGlobalConstDecl()" << std::endl;
#endif
    if (!constDecl) {
        return;
    }
    for (auto constDef : constDecl->const_defs->const_defs) {
        transGlobalConstDef(constDef, constDecl->type);
    }
}

void AST2IRP::transGlobalConstDef(ast::ConstDef *constDef, ast::BaseType baseType) {
    if (!constDef) {
        return;
    }
    if (!constDef->array_dim) {
        ExpTy *initVal = transExp(constDef->init_val->u.val);
        irp::Exp *initExp = unEx(initVal->exp);

        switch (baseType) {
        case ast::BaseType::INT: {
            initExp = irp::CastExp(initExp, irp::ExpType::INT);
            assert(initExp->kind == irp::ExpKind::CONSTI);
            this->globalTable.enter(*constDef->id, env::GlobalConstIntEntry(initExp->u.consti));
            break;
        }
        case ast::BaseType::FLOAT: {
            initExp = irp::CastExp(initExp, irp::ExpType::FLOAT);
            assert(initExp->kind == irp::ExpKind::CONSTF);
            this->globalTable.enter(*constDef->id, env::GlobalConstFloatEntry(initExp->u.constf));
            break;
        }
        default:
            assert(0);
        }
    } else {
        TrExpList *dims = transExpList(constDef->array_dim);
        irp::ExpList *irpDims = unTrExpList(dims);
        std::vector<int> *dimList = getArrayDimList(irpDims);

        temp::Label *label = temp::newNamedLabel(*constDef->id);

        switch (baseType) {
        case ast::BaseType::INT: {
            this->globalTable.enter(*constDef->id, env::GlobalConstIArrEntry(label, dimList));
            break;
        }
        case ast::BaseType::FLOAT: {
            this->globalTable.enter(*constDef->id, env::GlobalConstFArrEntry(label, dimList));
            break;
        }
        default:
            assert(0);
        }

        this->result->globalDefs.push_back(transGlobalArrConstInitVal(constDef->init_val, baseType, label, dimList));
    }
}

TrExp *AST2IRP::transGlobalVarDecl(ast::VarDecl *varDecl) {
#ifdef IRPGEN_DEBUG
    std::cerr << "AST2IRP::transGlobalVarDecl()" << std::endl;
#endif
    if (!varDecl) {
        return nullptr;
    }
    TrExp *res = nullptr;
    for (auto varDef : varDecl->var_defs->var_defs) {
        res = TrStmtList(res, transGlobalVarDef(varDef, varDecl->type));
    }
    return res;
}

TrExp *AST2IRP::transGlobalVarDef(ast::VarDef *varDef, ast::BaseType baseType) {
#ifdef IRPGEN_DEBUG
    std::cerr << "AST2IRP::transGlobalVarDef() for " << *varDef->id << std::endl;
#endif
    if (!varDef) {
        return nullptr;
    }
    temp::Label *label = temp::newNamedLabel(*varDef->id);
    this->globalTable.enter(*varDef->id, env::GlobalIntEntry(label));

    if (!varDef->array_dim) {
        auto [globalDef, initExp] = transGlobalVarInitVal(varDef->init_val, baseType, label);
        if (globalDef) {
            this->result->globalDefs.push_back(globalDef);
        }
        return initExp;
    } else {
        TrExpList *dims = transExpList(varDef->array_dim);
        irp::ExpList *irpDims = unTrExpList(dims);
        std::vector<int> *dimList = getArrayDimList(irpDims);

        switch (baseType) {
        case ast::BaseType::INT: {
            this->globalTable.enter(*varDef->id, env::GlobalIArrEntry(label, dimList));
            break;
        case ast::BaseType::FLOAT: {
            this->globalTable.enter(*varDef->id, env::GlobalFArrEntry(label, dimList));
            break;
        }
        default:
            assert(0);
        }
        }

        auto [globalDef, initExp] = transGlobalArrInitVal(varDef->init_val, baseType, label, dimList);
        if (globalDef) {
            this->result->globalDefs.push_back(globalDef);
        }
        return initExp;
    }
}

std::pair<irp::GlobalDef *, TrExp *> AST2IRP::transGlobalVarInitVal(ast::InitVal *initVal, ast::BaseType baseType,
                                                                    temp::Label *label) {
    switch (baseType) {
    case ast::BaseType::INT: {
        if (!initVal) {
            auto globalDef =
                irp::GlobalVarDef(label, new std::string(""), new std::string("\t.comm\t" + label->name + ",4,4\n"));
            return {globalDef, nullptr};
        } else {
            ExpTy *initExpTy = transExp(initVal->u.val);
            irp::Exp *initExp = irp::CastExp(unEx(initExpTy->exp), irp::ExpType::INT);
            if (initExp->kind == irp::ExpKind::CONSTI) {
                if (initExp->u.consti == 0) {
                    auto globalDef = irp::GlobalVarDef(label, new std::string(""),
                                                       new std::string("\t.global\t" + label->name +
                                                                       "\n"
                                                                       "\t.bss\n"
                                                                       "\t.align\t2\n"
                                                                       "\t.type\t" +
                                                                       label->name +
                                                                       ", %object\n"
                                                                       "\t.size\t" +
                                                                       label->name + ", 4\n" + label->name +
                                                                       ":\n"
                                                                       "\t.space\t4\n"));
                    return {globalDef, nullptr};
                } else {
                    auto globalDef = irp::GlobalVarDef(label, new std::string(""),
                                                       new std::string("\t.global\t" + label->name +
                                                                       "\n"
                                                                       "\t.data\n"
                                                                       "\t.align\t2\n"
                                                                       "\t.type\t" +
                                                                       label->name +
                                                                       ", %object\n"
                                                                       "\t.size\t" +
                                                                       label->name + ", 4\n" + label->name +
                                                                       ":\n"
                                                                       "\t.word\t" +
                                                                       std::to_string(initExp->u.consti) + "\n"));
                    return {globalDef, nullptr};
                }
            } else {
                auto globalDef = irp::GlobalVarDef(label, new std::string(""),
                                                   new std::string("\t.global\t" + label->name +
                                                                   "\n"
                                                                   "\t.data\n"
                                                                   "\t.align\t2\n"
                                                                   "\t.type\t" +
                                                                   label->name +
                                                                   ", %object\n"
                                                                   "\t.size\t" +
                                                                   label->name + ", 4\n" + label->name +
                                                                   ":\n"
                                                                   "\t.word\t0\n"));
                return {globalDef,
                        new TrExp(irp::MoveStmt(irp::MemExp(irp::NameExp(label), irp::ExpType::INT), initExp))};
            }
        }
    }
    case ast::BaseType::FLOAT: {
        if (!initVal) {
            auto globalDef =
                irp::GlobalVarDef(label, new std::string(""), new std::string("\t.comm\t" + label->name + ",4,4\n"));
            return {globalDef, nullptr};
        } else {
            ExpTy *initExpTy = transExp(initVal->u.val);
            irp::Exp *initExp = irp::CastExp(unEx(initExpTy->exp), irp::ExpType::FLOAT);
            if (initExp->kind == irp::ExpKind::CONSTF) {
                if (initExp->u.constf == 0.0) {
                    auto globalDef = irp::GlobalVarDef(label, new std::string(""),
                                                       new std::string("\t.global\t" + label->name +
                                                                       "\n"
                                                                       "\t.bss\n"
                                                                       "\t.align\t2\n"
                                                                       "\t.type\t" +
                                                                       label->name +
                                                                       ", %object\n"
                                                                       "\t.size\t" +
                                                                       label->name + ", 4\n" + label->name +
                                                                       ":\n"
                                                                       "\t.space\t4\n"));
                    return {globalDef, nullptr};
                } else {
                    auto globalDef =
                        irp::GlobalVarDef(label, new std::string(""),
                                          new std::string("\t.global\t" + label->name +
                                                          "\n"
                                                          "\t.data\n"
                                                          "\t.align\t2\n"
                                                          "\t.type\t" +
                                                          label->name +
                                                          ", %object\n"
                                                          "\t.size\t" +
                                                          label->name + ", 4\n" + label->name +
                                                          ":\n"
                                                          "\t.word\t" +
                                                          std::to_string(float2unsigned(initExp->u.constf)) + "\n"));
                    return {globalDef, nullptr};
                }
            } else {
                auto globalDef = irp::GlobalVarDef(label, new std::string(""),
                                                   new std::string("\t.global\t" + label->name +
                                                                   "\n"
                                                                   "\t.data\n"
                                                                   "\t.align\t2\n"
                                                                   "\t.type\t" +
                                                                   label->name +
                                                                   ", %object\n"
                                                                   "\t.size\t" +
                                                                   label->name + ", 4\n" + label->name +
                                                                   ":\n"
                                                                   "\t.word\t0\n"));
                return {globalDef,
                        new TrExp(irp::MoveStmt(irp::MemExp(irp::NameExp(label), irp::ExpType::FLOAT), initExp))};
            }
        }
    }
    default:
        assert(0);
    }
}

std::pair<irp::GlobalDef *, TrExp *> AST2IRP::transGlobalArrInitVal(ast::InitVal *initVal, ast::BaseType baseType,
                                                                    temp::Label *label, std::vector<int> *dims) {
#ifdef IRPGEN_DEBUG
    std::cerr << "AST2IRP::transGlobalArrInitVal()" << std::endl;
#endif
    int size = std::accumulate(dims->begin(), dims->end(), 1, std::multiplies<int>());
    if (!initVal) {
        return {irp::GlobalArrDef(label, new std::string(""),
                                  new std::string("\t.comm\t" + label->name + "," + std::to_string(4 * size) + ",4\n")),
                nullptr};
    }

    int offset = 0;
    auto recurResult = transGlobalInitValRecur(initVal, baseType, label, dims, 0, size, size, offset);

    std::string *llvmStr = new std::string("");
    std::string *armStr = new std::string("");

    llvmStr->append(*recurResult->llvmStr);

    armStr->append("\t.global\t" + label->name + "\n");
    armStr->append("\t.data\n");
    armStr->append("\t.align\t2\n");
    armStr->append("\t.type\t" + label->name + ", %object\n");
    armStr->append("\t.size\t" + label->name + ", " + std::to_string(4 * size) + "\n");
    armStr->append(label->name + ":\n");
    armStr->append(*recurResult->armStr);

#ifdef IRPGEN_DEBUG
    std::cerr << "armStr:\n" << *armStr << std::endl;
#endif

    return {irp::GlobalArrDef(label, llvmStr, armStr), recurResult->stmt};
}

InitValInfo *AST2IRP::transGlobalInitValRecur(ast::InitVal *initVal, ast::BaseType baseType, temp::Label *label,
                                              std::vector<int> *dims, int curDim, int step, int align, int &offset) {
#ifdef IRPGEN_DEBUG
    std::cerr << "AST2IRP::transGlobalInitValRecur(curDim=" << curDim << ", step=" << step << ", align=" << align
              << ", offset=" << offset << ")" << std::endl;
#endif
    switch (initVal->kind) {
    case ast::InitValKind::VAL: {
        ExpTy *val = transExp(initVal->u.val);
        assert(val->ty->isConst);
        irp::Exp *exp = unEx(val->exp);

        InitValInfo *res;
        switch (baseType) {
        case ast::BaseType::INT: {
            exp = irp::CastExp(exp, irp::ExpType::INT);
            if (exp->kind == irp::ExpKind::CONSTI) {
                if (exp->u.consti != 0) {
                    res = new InitValInfo(1, nullptr, new std::string(""),
                                          new std::string(util::format("\t.word\t%d\n", exp->u.consti)));
                } else {
                    res = new InitValInfo(1, nullptr, new std::string(""), new std::string("\t.space\t4\n"));
                }
            } else {
                irp::Exp *ptr = irp::BinOpExp(irp::BinOp::ADD, irp::NameExp(label), irp::ConstIExp(offset * 4));
                res = new InitValInfo(1, new TrExp(irp::MoveStmt(irp::MemExp(ptr, irp::ExpType::INT), exp)),
                                      new std::string(""), new std::string("\t.space\t4\n"));
            }
            break;
        }
        case ast::BaseType::FLOAT: {
            exp = irp::CastExp(exp, irp::ExpType::FLOAT);
            if (exp->kind == irp::ExpKind::CONSTF) {
                if (exp->u.constf != 0.0) {
                    res =
                        new InitValInfo(1, nullptr, new std::string(""),
                                        new std::string(util::format("\t.word\t%u\n", float2unsigned(exp->u.constf))));
                } else {
                    res = new InitValInfo(1, nullptr, new std::string(""), new std::string("\t.space\t4\n"));
                }
            } else {
                irp::Exp *ptr = irp::BinOpExp(irp::BinOp::ADD, irp::NameExp(label), irp::ConstIExp(offset * 4));
                res = new InitValInfo(1, new TrExp(irp::MoveStmt(irp::MemExp(ptr, irp::ExpType::FLOAT), exp)),
                                      new std::string(""), new std::string("\t.space\t4\n"));
            }
            break;
        }
        default:
            assert(0);
        }
        offset++;
        return res;
    }
    case ast::InitValKind::ARRAY: {
        if (!initVal->u.array) {
            auto res = new InitValInfo(align, nullptr, new std::string(""),
                                       new std::string(util::format("\t.space\t%d\n", align * 4)));
            offset += align;
            return res;
        }
        if (align < 0) {
            auto res = new InitValInfo(0, nullptr, new std::string(""), new std::string(""));
            for (auto val : initVal->u.array->array) {
                auto info = transGlobalInitValRecur(val, baseType, label, dims, curDim + 1, step / dims->at(curDim),
                                                    align, offset);
                res->size += info->size;
                res->stmt = TrStmtList(res->stmt, info->stmt);
                res->llvmStr->append(*info->llvmStr);
                res->armStr->append(*info->armStr);
            }
            return res;
        } else {
            auto res = new InitValInfo(0, nullptr, new std::string(""), new std::string(""));
            for (auto val : initVal->u.array->array) {
                int curAlign = -1;
                for (int i = curDim, curStep = step; i < dims->size() - 1; ++i) {
                    curStep /= dims->at(i);
                    if (offset % curStep == 0) {
                        curAlign = curStep;
                        break;
                    }
                }
                auto info = transGlobalInitValRecur(val, baseType, label, dims, curDim + 1, step / dims->at(curDim),
                                                    curAlign, offset);
                res->size += info->size;
                res->stmt = TrStmtList(res->stmt, info->stmt);
                res->llvmStr->append(*info->llvmStr);
                res->armStr->append(*info->armStr);
            }
            if (offset % align != 0) {
                res->size += align - offset % align;
                res->llvmStr->append("");
                res->armStr->append(util::format("\t.space\t%d\n", 4 * (align - offset % align)));
                offset += align - offset % align;
            }
            return res;
        }
    }
    default:
        assert(0);
    }
}

irp::GlobalDef *AST2IRP::transGlobalArrConstInitVal(ast::InitVal *initVal, ast::BaseType baseType, temp::Label *label,
                                                    std::vector<int> *dims) {
    int size = std::accumulate(dims->begin(), dims->end(), 1, std::multiplies<int>());
    int offset = 0;
    auto recurResult = transConstInitValRecur(initVal, baseType, dims, 0, size, size, offset);

    std::string *llvmStr = new std::string("");
    std::string *armStr = new std::string("");

    llvmStr->append(*recurResult->llvmStr);

    armStr->append("\t.global\t" + label->name + "\n");
    armStr->append("\t.section\t.rodata\n");
    armStr->append("\t.align\t2\n");
    armStr->append("\t.type\t" + label->name + ", %object\n");
    armStr->append("\t.size\t" + label->name + ", " + std::to_string(4 * size) + "\n");
    armStr->append(label->name + ":\n");
    armStr->append(*recurResult->armStr);

    return irp::GlobalConstArrDef(label, llvmStr, armStr);
}

ConstInitValInfo *AST2IRP::transConstInitValRecur(ast::InitVal *initVal, ast::BaseType baseType, std::vector<int> *dims,
                                                  int curDim, int step, int align, int &offset) {
    switch (initVal->kind) {
    case ast::InitValKind::VAL: {
        ExpTy *val = transExp(initVal->u.val);
        assert(val->ty->isConst);
        irp::Exp *exp = unEx(val->exp);

        ConstInitValInfo *res;
        switch (baseType) {
        case ast::BaseType::INT: {
            exp = irp::CastExp(exp, irp::ExpType::INT);
            if (exp->u.consti != 0) {
                res = new ConstInitValInfo(false, 1, new std::string(""),
                                           new std::string(util::format("\t.word\t%d\n", exp->u.consti)));
            } else {
                res = new ConstInitValInfo(true, 1, new std::string(""), new std::string("\t.space\t4\n"));
            }
            break;
        }
        case ast::BaseType::FLOAT: {
            exp = irp::CastExp(exp, irp::ExpType::FLOAT);
            if (exp->u.constf != 0.0) {
                res =
                    new ConstInitValInfo(false, 1, new std::string(""),
                                         new std::string(util::format("\t.word\t%u\n", float2unsigned(exp->u.constf))));
            } else {
                res = new ConstInitValInfo(true, 1, new std::string(""), new std::string("\t.space\t4\n"));
            }
            break;
        }
        default:
            assert(0);
        }
        offset++;
        return res;
    }
    case ast::InitValKind::ARRAY: {
        if (!initVal->u.array) {
            auto res = new ConstInitValInfo(true, align, new std::string(""),
                                            new std::string(util::format("\t.space\t%d\n", align * 4)));
            offset += align;
            return res;
        }
        if (align < 0) {
            auto res = new ConstInitValInfo(true, 0, new std::string(""), new std::string(""));
            for (auto val : initVal->u.array->array) {
                auto info =
                    transConstInitValRecur(val, baseType, dims, curDim + 1, step / dims->at(curDim), align, offset);
                res->isFilledZero &= info->isFilledZero;
                res->size += info->size;
                res->llvmStr->append(*info->llvmStr);
                res->armStr->append(*info->armStr);
            }
            return res;
        } else {
            auto res = new ConstInitValInfo(true, 0, new std::string(""), new std::string(""));
            for (auto val : initVal->u.array->array) {
                int curAlign = -1;
                for (int i = curDim, curStep = step; i < dims->size() - 1; ++i) {
                    curStep /= dims->at(i);
                    if (offset % curStep == 0) {
                        curAlign = curStep;
                        break;
                    }
                }
                auto info =
                    transConstInitValRecur(val, baseType, dims, curDim + 1, step / dims->at(curDim), curAlign, offset);
                res->isFilledZero &= info->isFilledZero;
                res->size += info->size;
                res->llvmStr->append(*info->llvmStr);
                res->armStr->append(*info->armStr);
            }
            if (offset % align != 0) {
                res->size += align - offset % align;
                res->llvmStr->append("");
                res->armStr->append(util::format("\t.space\t%d\n", 4 * (align - offset % align)));
                offset += align - offset % align;
            }
            return res;
        }
    }
    default:
        assert(0);
    }
}

TrExp *AST2IRP::transBlock(ast::Block *block) {
#ifdef IRPGEN_DEBUG
    std::cerr << "AST2IRP::transBlock()" << std::endl;
#endif
    if (!block) {
        return nullptr;
    }
    this->localTable.beginScope();
    auto res = transBlockItemList(block->block_item_list);
    this->localTable.endScope();
    return res;
}

TrExp *AST2IRP::transBlockItemList(ast::BlockItemList *blockItemList) {
    if (!blockItemList) {
        return nullptr;
    }
    TrExp *res = nullptr;
    for (auto blockItem : blockItemList->block_items) {
        TrExp *trExp = transBlockItem(blockItem);
        if (!trExp) {
            continue;
        }
        if (!res) {
            res = trExp;
        } else {
            res = new TrExp(irp::SeqStmt(unNx(res), unNx(trExp)));
        }
    }
    return res;
}

TrExp *AST2IRP::transBlockItem(ast::BlockItem *blockItem) {
    if (!blockItem) {
        return nullptr;
    }
    switch (blockItem->kind) {
    case ast::BlockItemKind::DECL: {
        return transDecl(blockItem->u.decl);
    }
    case ast::BlockItemKind::STMT: {
        return transStmt(blockItem->u.stmt);
    }
    default:
        assert(0);
    }
}

TrExp *AST2IRP::transDecl(ast::Decl *decl) {
    if (!decl) {
        return nullptr;
    }
    switch (decl->kind) {
    case ast::DeclKind::CONST: {
        transConstDecl(decl->u.const_decl);
        return nullptr;
    }
    case ast::DeclKind::VAR: {
        return transVarDecl(decl->u.var_decl);
    }
    default:
        assert(0);
    }
}

void AST2IRP::transConstDecl(ast::ConstDecl *constDecl) {
    if (!constDecl) {
        return;
    }
    for (auto constDef : constDecl->const_defs->const_defs) {
        transConstDef(constDef, constDecl->type);
    }
}

void AST2IRP::transConstDef(ast::ConstDef *constDef, ast::BaseType baseType) {
    if (!constDef) {
        return;
    }
    if (!constDef->array_dim) {
        ExpTy *initVal = transExp(constDef->init_val->u.val);
        irp::Exp *initExp = unEx(initVal->exp);

        switch (baseType) {
        case ast::BaseType::INT: {
            initExp = irp::CastExp(initExp, irp::ExpType::INT);
            assert(initExp->kind == irp::ExpKind::CONSTI);
            this->localTable.enter(*constDef->id, env::LocalConstIntEntry(initExp->u.consti));
            break;
        }
        case ast::BaseType::FLOAT: {
            initExp = irp::CastExp(initExp, irp::ExpType::FLOAT);
            assert(initExp->kind == irp::ExpKind::CONSTF);
            this->localTable.enter(*constDef->id, env::LocalConstFloatEntry(initExp->u.constf));
            break;
        }
        default:
            assert(0);
        }
    } else {
        std::string assemId = *constDef->id + "." + *this->curFuncName;
        temp::Label *label = temp::newNamedLabel(assemId);

        TrExpList *dims = transExpList(constDef->array_dim);
        irp::ExpList *irpDims = unTrExpList(dims);
        std::vector<int> *dimList = getArrayDimList(irpDims);

        switch (baseType) {
        case ast::BaseType::INT: {
            this->localTable.enter(*constDef->id, env::LocalConstIArrEntry(label, dimList));
            break;
        }
        case ast::BaseType::FLOAT: {
            this->localTable.enter(*constDef->id, env::LocalConstFArrEntry(label, dimList));
            break;
        }
        default:
            assert(0);
        }

        this->result->localConstArrayDefs.push_back(
            transLocalArrConstInitVal(constDef->init_val, baseType, label, dimList));
    }
}

TrExp *AST2IRP::transVarDecl(ast::VarDecl *varDecl) {
    if (!varDecl) {
        return nullptr;
    }
    TrExp *res = nullptr;
    for (auto varDef : varDecl->var_defs->var_defs) {
        res = TrStmtList(res, transVarDef(varDef, varDecl->type));
    }
    return res;
}

TrExp *AST2IRP::transVarDef(ast::VarDef *varDef, ast::BaseType baseType) {
    if (!varDef) {
        return nullptr;
    }
    if (!varDef->array_dim) {
        switch (baseType) {
        case ast::BaseType::INT: {
            temp::Temp *temp = temp::newTemp(temp::TempType::INT);
            this->localTable.enter(*varDef->id, env::LocalIntEntry(temp));
            if (!varDef->init_val) {
                return nullptr;
            } else {
                ExpTy *initVal = transExp(varDef->init_val->u.val);
                irp::Exp *initExp = unEx(initVal->exp);
                if (initExp->type != irp::ExpType::INT) {
                    initExp = irp::CastExp(initExp, irp::ExpType::INT);
                }
                return new TrExp(irp::MoveStmt(irp::TempExp(temp), initExp));
            }
        }
        case ast::BaseType::FLOAT: {
            temp::Temp *temp = temp::newTemp(temp::TempType::FLOAT);
            this->localTable.enter(*varDef->id, env::LocalFloatEntry(temp));
            if (!varDef->init_val) {
                return nullptr;
            } else {
                ExpTy *initVal = transExp(varDef->init_val->u.val);
                irp::Exp *initExp = unEx(initVal->exp);
                if (initExp->type != irp::ExpType::FLOAT) {
                    initExp = irp::CastExp(initExp, irp::ExpType::FLOAT);
                }
                return new TrExp(irp::MoveStmt(irp::TempExp(temp), initExp));
            }
        }
        default:
            assert(0);
        }
    } else {
        TrExpList *dims = transExpList(varDef->array_dim);
        irp::ExpList *irpDims = unTrExpList(dims);
        std::vector<int> *dimList = getArrayDimList(irpDims);
        int size = std::accumulate(dimList->begin(), dimList->end(), 1, std::multiplies<int>());
        irp::Exp *allocaExp = irp::AllocaExp(size, this->curFuncStackOffset);
        this->curFuncStackOffset += size;

        temp::Temp *temp = nullptr;
        switch (baseType) {
        case ast::BaseType::INT: {
            temp = temp::newTemp(temp::TempType::INT_PTR);
            break;
        }
        case ast::BaseType::FLOAT: {
            temp = temp::newTemp(temp::TempType::FLOAT_PTR);
            break;
        }
        default:
            assert(0);
        }
        this->localTable.enter(*varDef->id, env::LocalIArrEntry(temp, dimList));
        this->curFuncAllocas =
            TrStmtList(this->curFuncAllocas, new TrExp(irp::MoveStmt(irp::TempExp(temp), allocaExp)));
        return transInitVal(varDef->init_val, baseType, temp, dimList);
    }
}

TrExp *AST2IRP::transInitVal(ast::InitVal *initVal, ast::BaseType baseType, temp::Temp *temp, std::vector<int> *dims) {
    if (!initVal) {
        return nullptr;
    }
    int size = std::accumulate(dims->begin(), dims->end(), 1, std::multiplies<int>());
    int offset = 0;
    return transInitValRecur(initVal, baseType, temp, dims, 0, size, size, offset);
}

TrExp *AST2IRP::transInitValRecur(ast::InitVal *initVal, ast::BaseType baseType, temp::Temp *temp,
                                  std::vector<int> *dims, int curDim, int step, int align, int &offset) {
    switch (initVal->kind) {
    case ast::InitValKind::VAL: {
        ExpTy *val = transExp(initVal->u.val);
        assert(val->ty->isConst);
        irp::Exp *exp = unEx(val->exp);

        TrExp *res;
        switch (baseType) {
        case ast::BaseType::INT: {
            exp = irp::CastExp(exp, irp::ExpType::INT);
            res = new TrExp(irp::MoveStmt(
                irp::MemExp(irp::BinOpExp(irp::BinOp::ADD, irp::TempExp(temp), irp::ConstIExp(offset * 4)),
                            irp::ExpType::INT),
                exp));
            break;
        }
        case ast::BaseType::FLOAT: {
            exp = irp::CastExp(exp, irp::ExpType::FLOAT);
            res = new TrExp(irp::MoveStmt(
                irp::MemExp(irp::BinOpExp(irp::BinOp::ADD, irp::TempExp(temp), irp::ConstIExp(offset * 4)),
                            irp::ExpType::FLOAT),
                exp));
            break;
        }
        default:
            assert(0);
        }
        offset++;
        return res;
    }
    case ast::InitValKind::ARRAY: {
        if (!initVal->u.array) {
            TrExp *res = new TrExp(irp::MemsetStmt(temp, offset, align));
            offset += align;
            return res;
        }
        if (align < 0) {
            TrExp *res = nullptr;
            for (auto val : initVal->u.array->array) {
                res = TrStmtList(res, transInitValRecur(val, baseType, temp, dims, curDim + 1, step / dims->at(curDim),
                                                        align, offset));
            }
            return res;
        } else {
            TrExp *res = nullptr;
            for (auto val : initVal->u.array->array) {
                int curAlign = -1;
                for (int i = curDim, curStep = step; i < dims->size() - 1; ++i) {
                    curStep /= dims->at(i);
                    if (offset % curStep == 0) {
                        curAlign = curStep;
                        break;
                    }
                }
                res = TrStmtList(res, transInitValRecur(val, baseType, temp, dims, curDim + 1, step / dims->at(curDim),
                                                        curAlign, offset));
            }
            if (offset % align != 0) {
                res = TrStmtList(res, new TrExp(irp::MemsetStmt(temp, offset, align - offset % align)));
                offset += align - offset % align;
            }
            return res;
        }
    }
    }
}

irp::LocalConstArrayDef *AST2IRP::transLocalArrConstInitVal(ast::InitVal *initVal, ast::BaseType baseType,
                                                            temp::Label *label, std::vector<int> *dims) {
    int size = std::accumulate(dims->begin(), dims->end(), 1, std::multiplies<int>());
    int offset = 0;
    auto recurResult = transConstInitValRecur(initVal, baseType, dims, 0, size, size, offset);

    std::string *llvmStr = new std::string("");
    std::string *armStr = new std::string("");

    llvmStr->append(*recurResult->llvmStr);

    armStr->append("\t.section\t.rodata\n");
    armStr->append("\t.align\t2\n");
    armStr->append("\t.type\t" + label->name + ", %object\n");
    armStr->append("\t.size\t" + label->name + ", " + std::to_string(4 * size) + "\n");
    armStr->append(label->name + ":\n");
    armStr->append(*recurResult->armStr);

    return irp::LocalConstArrDef(label, llvmStr, armStr);
}

TrExp *AST2IRP::transStmt(ast::Stmt *stmt) {
    if (!stmt) {
        return nullptr;
    }
    switch (stmt->kind) {
    case ast::StmtKind::ASSIGN: {
        return transAssignStmt(stmt);
    }
    case ast::StmtKind::EXP: {
        return transExpStmt(stmt);
    }
    case ast::StmtKind::BLOCK: {
        return transBlockStmt(stmt);
    }
    case ast::StmtKind::IF: {
        return transIfStmt(stmt);
    }
    case ast::StmtKind::WHILE: {
        return transWhileStmt(stmt);
    }
    case ast::StmtKind::BREAK: {
        return transBreakStmt(stmt);
    }
    case ast::StmtKind::CONTINUE: {
        return transContinueStmt(stmt);
    }
    case ast::StmtKind::RETURN: {
        return transReturnStmt(stmt);
    }
    case ast::StmtKind::PUTINT: {
        return transPutintStmt(stmt);
    }
    case ast::StmtKind::PUTCH: {
        return transPutchStmt(stmt);
    }
    case ast::StmtKind::PUTFLOAT: {
        return transPutfloatStmt(stmt);
    }
    case ast::StmtKind::PUTARRAY: {
        return transPutarrayStmt(stmt);
    }
    case ast::StmtKind::PUTFARRAY: {
        return transPutfarrayStmt(stmt);
    }
    case ast::StmtKind::PUTF: {
        return transPutfStmt(stmt);
    }
    case ast::StmtKind::STARTTIME: {
        return transStarttimeStmt();
    }
    case ast::StmtKind::STOPTIME: {
        return transStoptimeStmt();
    }
    default:
        assert(0);
    }
}

TrExp *AST2IRP::transAssignStmt(ast::Stmt *stmt) {
    if (!stmt) {
        return nullptr;
    }
    ExpTy *left = transExp(stmt->u.assign.lhs);
    if (!left) {
        return nullptr;
    }
    ExpTy *right = transExp(stmt->u.assign.rhs);
    if (!right) {
        return nullptr;
    }

    irp::Exp *dst = unEx(left->exp);
    irp::Exp *src = unEx(right->exp);
    if (dst->type != src->type) {
        return new TrExp(irp::MoveStmt(dst, irp::CastExp(src, dst->type)));
    }
    return new TrExp(irp::MoveStmt(dst, src));
}

TrExp *AST2IRP::transExpStmt(ast::Stmt *stmt) {
    if (!stmt) {
        return nullptr;
    }
    ExpTy *exp = transExp(stmt->u.exp);
    if (!exp) {
        return nullptr;
    }
    return new TrExp(irp::ExpStmt(unEx(exp->exp)));
}

TrExp *AST2IRP::transBlockStmt(ast::Stmt *stmt) {
    if (!stmt) {
        return nullptr;
    }
    return transBlock(stmt->u.block);
}

TrExp *AST2IRP::transIfStmt(ast::Stmt *stmt) {
    if (!stmt) {
        return nullptr;
    }

    ExpTy *cond = transExp(stmt->u.if_stmt.cond);
    if (!cond) {
        return nullptr;
    }
    bool isCondConst = cond->ty->isConst;

    TrExp *thenStmt = transStmt(stmt->u.if_stmt.thenStmt);
    TrExp *elseStmt = transStmt(stmt->u.if_stmt.elseStmt);
    if (isCondConst) {
        if (cond->exp->kind != TrExp::TrEx ||
            (cond->exp->u.ex->kind != irp::ExpKind::CONSTI && cond->exp->u.ex->kind != irp::ExpKind::CONSTF)) {
            std::cerr << "AST2IRP::transIfStmt(): condition is not a constant" << std::endl;
            std::exit(1);
        }
        if (cond->exp->u.ex->kind == irp::ExpKind::CONSTI) {
            if (cond->exp->u.ex->u.consti == 0) {
                return elseStmt ? new TrExp(unNx(elseStmt)) : nullptr;
            }
            return thenStmt ? new TrExp(unNx(thenStmt)) : nullptr;
        } else {
            if (cond->exp->u.ex->u.constf == 0.0) {
                return elseStmt ? new TrExp(unNx(elseStmt)) : nullptr;
            }
            return thenStmt ? new TrExp(unNx(thenStmt)) : nullptr;
        }
    }

    if (!thenStmt && !elseStmt) {
#ifdef IRPGEN_DEBUG
        std::cerr << "AST2IRP::transIfStmt(): only need to evaluate the condition" << std::endl;
#endif
        if (isCondConst) {
            return nullptr;
        }
        // only need to evaluate the condition
        return new TrExp(irp::ExpStmt(unEx(cond->exp)));
    }
    if (!thenStmt) {
        // if (cond) else { ... }
        temp::Label *e = temp::newLabel();
        temp::Label *f = temp::newLabel();

        TrCx *cx = unCx(cond->exp);
        doPatch(cx->trues, e);
        doPatch(cx->falses, f);

        return new TrExp(
            irp::SeqStmt(cx->stmt, irp::SeqStmt(irp::LabelStmt(f), irp::SeqStmt(unNx(elseStmt), irp::LabelStmt(e)))));
    }
    if (!elseStmt) {
        // if (cond) { ... }
        temp::Label *t = temp::newLabel();
        temp::Label *e = temp::newLabel();

        TrCx *cx = unCx(cond->exp);
        doPatch(cx->trues, t);
        doPatch(cx->falses, e);

        return new TrExp(
            irp::SeqStmt(cx->stmt, irp::SeqStmt(irp::LabelStmt(t), irp::SeqStmt(unNx(thenStmt), irp::LabelStmt(e)))));
    }
    temp::Label *t = temp::newLabel();
    temp::Label *f = temp::newLabel();
    temp::Label *e = temp::newLabel();

    TrCx *cx = unCx(cond->exp);
    doPatch(cx->trues, t);
    doPatch(cx->falses, f);

    return new TrExp(irp::SeqStmt(
        cx->stmt,
        irp::SeqStmt(irp::LabelStmt(t),
                     irp::SeqStmt(unNx(thenStmt),
                                  irp::SeqStmt(irp::JumpStmt(e),
                                               irp::SeqStmt(irp::LabelStmt(f),
                                                            irp::SeqStmt(unNx(elseStmt), irp::LabelStmt(e))))))));
}

TrExp *AST2IRP::transWhileStmt(ast::Stmt *stmt) {
    if (!stmt) {
        return nullptr;
    }

    ExpTy *cond = transExp(stmt->u.while_stmt.cond);
    if (!cond) {
        return nullptr;
    }
    bool isCondConst = cond->ty->isConst;
    if (isCondConst) {
        if (cond->exp->kind != TrExp::TrEx ||
            (cond->exp->u.ex->kind != irp::ExpKind::CONSTI && cond->exp->u.ex->kind != irp::ExpKind::CONSTF)) {
            return nullptr;
        }
        if (cond->exp->u.ex->kind == irp::ExpKind::CONSTI) {
            if (cond->exp->u.ex->u.consti == 0) {
                return nullptr;
            }
        } else {
            if (cond->exp->u.ex->u.constf == 0.0) {
                return nullptr;
            }
        }
    }

    temp::Label *test = temp::newLabel();
    temp::Label *done = temp::newLabel();
    this->loopStack.push(new LoopInfo(test, done));

    TrExp *bodyStmt = transStmt(stmt->u.while_stmt.body);
    TrExp *result = nullptr;
    if (!bodyStmt) {
        TrCx *cx = unCx(cond->exp);
        doPatch(cx->trues, test);
        doPatch(cx->falses, done);
        result = new TrExp(irp::SeqStmt(irp::LabelStmt(test), irp::SeqStmt(cx->stmt, irp::LabelStmt(done))));
    } else {
        temp::Label *body = temp::newLabel();
        TrCx *cx = unCx(cond->exp);
        doPatch(cx->trues, body);
        doPatch(cx->falses, done);
        result = new TrExp(irp::SeqStmt(
            irp::LabelStmt(test),
            irp::SeqStmt(cx->stmt, irp::SeqStmt(irp::LabelStmt(body),
                                                irp::SeqStmt(unNx(bodyStmt), irp::SeqStmt(irp::JumpStmt(test),
                                                                                          irp::LabelStmt(done)))))));
    }

    this->loopStack.pop();
    return result;
}

TrExp *AST2IRP::transBreakStmt(ast::Stmt *stmt) {
    if (!stmt) {
        return nullptr;
    }
    return new TrExp(irp::JumpStmt(this->loopStack.top()->done));
}

TrExp *AST2IRP::transContinueStmt(ast::Stmt *stmt) {
    if (!stmt) {
        return nullptr;
    }
    return new TrExp(irp::JumpStmt(this->loopStack.top()->test));
}

TrExp *AST2IRP::transReturnStmt(ast::Stmt *stmt) {
    if (!stmt) {
        return nullptr;
    }
    ExpTy *exp = transExp(stmt->u.exp);
    if (!exp) {
        return nullptr;
    }
    auto func = this->funcTable.look(*this->curFuncName);
    switch (func->retType) {
    case env::FuncRetType::INT: {
        return new TrExp(irp::ReturnStmt(irp::CastExp(unEx(exp->exp), irp::ExpType::INT)));
    }
    case env::FuncRetType::FLOAT: {
        return new TrExp(irp::ReturnStmt(irp::CastExp(unEx(exp->exp), irp::ExpType::FLOAT)));
    }
    default:
        assert(0);
    }

    return nullptr;
}

TrExp *AST2IRP::transPutintStmt(ast::Stmt *stmt) {
    if (!stmt) {
        return nullptr;
    }
    ExpTy *exp = transExp(stmt->u.exp);
    if (!exp) {
        return nullptr;
    }
    return new TrExp(irp::PutintStmt(unEx(exp->exp)));
}

TrExp *AST2IRP::transPutchStmt(ast::Stmt *stmt) {
    if (!stmt) {
        return nullptr;
    }
    ExpTy *exp = transExp(stmt->u.exp);
    if (!exp) {
        return nullptr;
    }
    return new TrExp(irp::PutchStmt(unEx(exp->exp)));
}

TrExp *AST2IRP::transPutfloatStmt(ast::Stmt *stmt) {
    if (!stmt) {
        return nullptr;
    }
    ExpTy *exp = transExp(stmt->u.exp);
    if (!exp) {
        return nullptr;
    }
    return new TrExp(irp::PutfloatStmt(unEx(exp->exp)));
}

TrExp *AST2IRP::transPutarrayStmt(ast::Stmt *stmt) {
    if (!stmt) {
        return nullptr;
    }
    ExpTy *array = transExp(stmt->u.putarray.array);
    if (!array) {
        return nullptr;
    }
    ExpTy *outlen = transExp(stmt->u.putarray.out_len);
    if (!outlen) {
        return nullptr;
    }
    return new TrExp(irp::PutarrayStmt(unEx(array->exp), unEx(outlen->exp)));
}

TrExp *AST2IRP::transPutfarrayStmt(ast::Stmt *stmt) {
    if (!stmt) {
        return nullptr;
    }
    ExpTy *array = transExp(stmt->u.putarray.array);
    if (!array) {
        return nullptr;
    }
    ExpTy *outlen = transExp(stmt->u.putarray.out_len);
    if (!outlen) {
        return nullptr;
    }
    return new TrExp(irp::PutfarrayStmt(unEx(array->exp), unEx(outlen->exp)));
}

TrExp *AST2IRP::transPutfStmt(ast::Stmt *stmt) {
    if (!stmt) {
        return nullptr;
    }
    TrExpList *args = transExpList(stmt->u.putf.args);
    irp::ExpList *irpArgs = unTrExpList(args);
    return new TrExp(irp::PutfStmt(stmt->u.putf.fmt, irpArgs));
}

TrExp *AST2IRP::transStarttimeStmt() { return new TrExp(irp::StarttimeStmt()); }

TrExp *AST2IRP::transStoptimeStmt() { return new TrExp(irp::StoptimeStmt()); }

ExpTy *AST2IRP::transExp(ast::Exp *exp) {
    if (!exp) {
        return nullptr;
    }
    switch (exp->kind) {
    case ast::ExpKind::BINOP: {
        return transBinOpExp(exp);
    }
    case ast::ExpKind::COND: {
        return transCondExp(exp);
    }
    case ast::ExpKind::ID: {
        return transIdExp(exp);
    }
    case ast::ExpKind::INTCONST: {
        return transIntConstExp(exp);
    }
    case ast::ExpKind::FLOATCONST: {
        return transFloatConstExp(exp);
    }
    case ast::ExpKind::NOT: {
        return transNotExp(exp);
    }
    case ast::ExpKind::MINUS: {
        return transMinusExp(exp);
    }
    case ast::ExpKind::CALL: {
        return transCallExp(exp);
    }
    case ast::ExpKind::ARRAY: {
        return transArrayExp(exp);
    }
    case ast::ExpKind::GETINT: {
        return transGetintExp();
    }
    case ast::ExpKind::GETCH: {
        return transGetchExp();
    }
    case ast::ExpKind::GETFLOAT: {
        return transGetfloatExp();
    }
    case ast::ExpKind::GETARRAY: {
        return transGetarrayExp(exp);
    }
    case ast::ExpKind::GETFARRAY: {
        return transGetfarrayExp(exp);
    }
    default:
        assert(0);
    }
}

ExpTy *AST2IRP::transBinOpExp(ast::Exp *exp) {
    if (!exp) {
        return nullptr;
    }
    ExpTy *left = transExp(exp->u.binop.lhs);
    if (!left) {
        return nullptr;
    }
    ExpTy *right = transExp(exp->u.binop.rhs);
    if (!right) {
        return nullptr;
    }

    irp::Exp *result;
    switch (exp->u.binop.op) {
    case ast::BinOp::ADD: {
        result = irp::BinOpExp(irp::BinOp::ADD, unEx(left->exp), unEx(right->exp));
        break;
    }
    case ast::BinOp::SUB: {
        result = irp::BinOpExp(irp::BinOp::SUB, unEx(left->exp), unEx(right->exp));
        break;
    }
    case ast::BinOp::MUL: {
        result = irp::BinOpExp(irp::BinOp::MUL, unEx(left->exp), unEx(right->exp));
        break;
    }
    case ast::BinOp::DIV: {
        result = irp::BinOpExp(irp::BinOp::DIV, unEx(left->exp), unEx(right->exp));
        break;
    }
    case ast::BinOp::MOD: {
        result = irp::BinOpExp(irp::BinOp::MOD, unEx(left->exp), unEx(right->exp));
        break;
    }
    default:
        assert(0);
    }
    env::TyKind resType = result->type == irp::ExpType::INT ? env::TyKind::INT : env::TyKind::FLOAT;
    bool isConst = left->ty->isConst && right->ty->isConst;

    return new ExpTy(new env::Type(resType, isConst), new TrExp(result));
}

ExpTy *AST2IRP::transCondExp(ast::Exp *exp) {
    if (!exp) {
        return nullptr;
    }
    ExpTy *left = transExp(exp->u.cond.lhs);
    if (!left) {
        return nullptr;
    }
    ExpTy *right = transExp(exp->u.cond.rhs);
    if (!right) {
        return nullptr;
    }

    if (left->ty->isConst && right->ty->isConst) {
        irp::Exp *leftExp = unEx(left->exp);
        irp::Exp *rightExp = unEx(right->exp);
        if (leftExp->type == irp::ExpType::INT && rightExp->type == irp::ExpType::INT) {
            switch (exp->u.cond.op) {
            case ast::RelOp::GT:
                return new ExpTy(new env::Type(env::TyKind::INT, true),
                                 new TrExp(irp::ConstIExp(leftExp->u.consti > rightExp->u.consti)));
            case ast::RelOp::LT:
                return new ExpTy(new env::Type(env::TyKind::INT, true),
                                 new TrExp(irp::ConstIExp(leftExp->u.consti < rightExp->u.consti)));
            case ast::RelOp::GE:
                return new ExpTy(new env::Type(env::TyKind::INT, true),
                                 new TrExp(irp::ConstIExp(leftExp->u.consti >= rightExp->u.consti)));
            case ast::RelOp::LE:
                return new ExpTy(new env::Type(env::TyKind::INT, true),
                                 new TrExp(irp::ConstIExp(leftExp->u.consti <= rightExp->u.consti)));
            case ast::RelOp::EQ:
                return new ExpTy(new env::Type(env::TyKind::INT, true),
                                 new TrExp(irp::ConstIExp(leftExp->u.consti == rightExp->u.consti)));
            case ast::RelOp::NE:
                return new ExpTy(new env::Type(env::TyKind::INT, true),
                                 new TrExp(irp::ConstIExp(leftExp->u.consti != rightExp->u.consti)));
            case ast::RelOp::AND:
                return new ExpTy(new env::Type(env::TyKind::INT, true),
                                 new TrExp(irp::ConstIExp(leftExp->u.consti && rightExp->u.consti)));
            case ast::RelOp::OR:
                return new ExpTy(new env::Type(env::TyKind::INT, true),
                                 new TrExp(irp::ConstIExp(leftExp->u.consti || rightExp->u.consti)));
            default:
                assert(0);
            }
        } else {
            float lv = leftExp->type == irp::ExpType::INT ? static_cast<float>(leftExp->u.consti) : leftExp->u.constf;
            float rv =
                rightExp->type == irp::ExpType::INT ? static_cast<float>(rightExp->u.consti) : rightExp->u.constf;
            switch (exp->u.cond.op) {
            case ast::RelOp::GT:
                return new ExpTy(new env::Type(env::TyKind::INT, true), new TrExp(irp::ConstIExp(lv > rv)));
            case ast::RelOp::LT:
                return new ExpTy(new env::Type(env::TyKind::INT, true), new TrExp(irp::ConstIExp(lv < rv)));
            case ast::RelOp::GE:
                return new ExpTy(new env::Type(env::TyKind::INT, true), new TrExp(irp::ConstIExp(lv >= rv)));
            case ast::RelOp::LE:
                return new ExpTy(new env::Type(env::TyKind::INT, true), new TrExp(irp::ConstIExp(lv <= rv)));
            case ast::RelOp::EQ:
                return new ExpTy(new env::Type(env::TyKind::INT, true), new TrExp(irp::ConstIExp(lv == rv)));
            case ast::RelOp::NE:
                return new ExpTy(new env::Type(env::TyKind::INT, true), new TrExp(irp::ConstIExp(lv != rv)));
            case ast::RelOp::AND:
                return new ExpTy(new env::Type(env::TyKind::INT, true), new TrExp(irp::ConstIExp(lv && rv)));
            case ast::RelOp::OR:
                return new ExpTy(new env::Type(env::TyKind::INT, true), new TrExp(irp::ConstIExp(lv || rv)));
            default:
                assert(0);
            }
        }
    }

    switch (exp->u.cond.op) {
    case ast::RelOp::GT: {
        irp::Stmt *cjump = irp::CJumpStmt(irp::RelOp::GT, unEx(left->exp), unEx(right->exp), nullptr, nullptr);
        PatchList *trues = new PatchList(&(cjump->u.cjump.t), nullptr);
        PatchList *falses = new PatchList(&(cjump->u.cjump.f), nullptr);
        return new ExpTy(new env::Type(env::TyKind::VOID, false), new TrExp(trues, falses, cjump));
    }
    case ast::RelOp::LT: {
        irp::Stmt *cjump = irp::CJumpStmt(irp::RelOp::LT, unEx(left->exp), unEx(right->exp), nullptr, nullptr);
        PatchList *trues = new PatchList(&(cjump->u.cjump.t), nullptr);
        PatchList *falses = new PatchList(&(cjump->u.cjump.f), nullptr);
        return new ExpTy(new env::Type(env::TyKind::VOID, false), new TrExp(trues, falses, cjump));
    }
    case ast::RelOp::GE: {
        irp::Stmt *cjump = irp::CJumpStmt(irp::RelOp::GE, unEx(left->exp), unEx(right->exp), nullptr, nullptr);
        PatchList *trues = new PatchList(&(cjump->u.cjump.t), nullptr);
        PatchList *falses = new PatchList(&(cjump->u.cjump.f), nullptr);
        return new ExpTy(new env::Type(env::TyKind::VOID, false), new TrExp(trues, falses, cjump));
    }
    case ast::RelOp::LE: {
        irp::Stmt *cjump = irp::CJumpStmt(irp::RelOp::LE, unEx(left->exp), unEx(right->exp), nullptr, nullptr);
        PatchList *trues = new PatchList(&(cjump->u.cjump.t), nullptr);
        PatchList *falses = new PatchList(&(cjump->u.cjump.f), nullptr);
        return new ExpTy(new env::Type(env::TyKind::VOID, false), new TrExp(trues, falses, cjump));
    }
    case ast::RelOp::EQ: {
        irp::Stmt *cjump = irp::CJumpStmt(irp::RelOp::EQ, unEx(left->exp), unEx(right->exp), nullptr, nullptr);
        PatchList *trues = new PatchList(&(cjump->u.cjump.t), nullptr);
        PatchList *falses = new PatchList(&(cjump->u.cjump.f), nullptr);
        return new ExpTy(new env::Type(env::TyKind::VOID, false), new TrExp(trues, falses, cjump));
    }
    case ast::RelOp::NE: {
        irp::Stmt *cjump = irp::CJumpStmt(irp::RelOp::NE, unEx(left->exp), unEx(right->exp), nullptr, nullptr);
        PatchList *trues = new PatchList(&(cjump->u.cjump.t), nullptr);
        PatchList *falses = new PatchList(&(cjump->u.cjump.f), nullptr);
        return new ExpTy(new env::Type(env::TyKind::VOID, false), new TrExp(trues, falses, cjump));
    }
    case ast::RelOp::AND: {
        /*
        left && right
        <=>
            Cjump(T_ne, unEx(left), 0, rightlabel, leftf)
        rightlabel:
            Cjump(T_ne, unEx(right), 0, rightt, rightf)
        t:
            ...
        f:
            ...
        [constraint]
            t = rightt
            f = leftf, rightf
        */
        if (left->ty->isConst) {
            if (left->exp->kind != TrExp::TrEx ||
                (left->exp->u.ex->kind != irp::ExpKind::CONSTI && left->exp->u.ex->kind != irp::ExpKind::CONSTF)) {
                std::cerr << "AST2IRP::transCondExp(): left operand is not a constant" << std::endl;
                std::exit(1);
            }
            bool isLeftTrue = left->exp->u.ex->kind == irp::ExpKind::CONSTI ? left->exp->u.ex->u.consti != 0
                                                                            : left->exp->u.ex->u.constf != 0.0;
            if (!isLeftTrue) {
                return new ExpTy(new env::Type(env::TyKind::INT, true), new TrExp(irp::ConstIExp(0)));
            } else {
                TrCx *rightCx = unCx(right->exp);
                return new ExpTy(new env::Type(env::TyKind::VOID, false),
                                 new TrExp(rightCx->trues, rightCx->falses, rightCx->stmt));
            }
        } else if (right->ty->isConst) {
            if (right->exp->kind != TrExp::TrEx ||
                (right->exp->u.ex->kind != irp::ExpKind::CONSTI && right->exp->u.ex->kind != irp::ExpKind::CONSTF)) {
                std::cerr << "AST2IRP::transCondExp(): right operand is not a constant" << std::endl;
                std::exit(1);
            }
            bool isRightTrue = right->exp->u.ex->kind == irp::ExpKind::CONSTI ? right->exp->u.ex->u.consti != 0
                                                                              : right->exp->u.ex->u.constf != 0.0;
            if (!isRightTrue) {
                return new ExpTy(new env::Type(env::TyKind::INT, true), new TrExp(irp::ConstIExp(0)));
            } else {
                TrCx *leftCx = unCx(left->exp);
                return new ExpTy(new env::Type(env::TyKind::VOID, false),
                                 new TrExp(leftCx->trues, leftCx->falses, leftCx->stmt));
            }
        } else {
            TrCx *leftCx = unCx(left->exp);
            TrCx *rightCx = unCx(right->exp);

            temp::Label *rightLabel = temp::newLabel();
            doPatch(leftCx->trues, rightLabel);

            PatchList *trues = rightCx->trues;
            PatchList *falses = joinPatch(leftCx->falses, rightCx->falses);

            return new ExpTy(
                new env::Type(env::TyKind::INT, false),
                new TrExp(trues, falses,
                          irp::SeqStmt(leftCx->stmt, irp::SeqStmt(irp::LabelStmt(rightLabel), rightCx->stmt))));
        }
    }
    case ast::RelOp::OR: {
        /*
        left || right
        <=>
            Cjump(T_ne, unEx(left), 0, leftt, rightlabel)
        rightlabel:
            Cjump(T_ne, unEx(right), 0, rightt, rightf)
        t:
            ...
        f:
            ...
        [constraint]
            t = leftt, rightt
            f = rightf
        */
        if (left->ty->isConst) {
            if (left->exp->kind != TrExp::TrEx ||
                (left->exp->u.ex->kind != irp::ExpKind::CONSTI && left->exp->u.ex->kind != irp::ExpKind::CONSTF)) {
                std::cerr << "AST2IRP::transCondExp(): left operand is not a constant" << std::endl;
                std::exit(1);
            }
            bool isLeftTrue = left->exp->u.ex->kind == irp::ExpKind::CONSTI ? left->exp->u.ex->u.consti != 0
                                                                            : left->exp->u.ex->u.constf != 0.0;
            if (isLeftTrue) {
                return new ExpTy(new env::Type(env::TyKind::INT, true), new TrExp(irp::ConstIExp(1)));
            } else {
                TrCx *rightCx = unCx(right->exp);
                return new ExpTy(new env::Type(env::TyKind::VOID, false),
                                 new TrExp(rightCx->trues, rightCx->falses, rightCx->stmt));
            }
        } else if (right->ty->isConst) {
            if (right->exp->kind != TrExp::TrEx ||
                (right->exp->u.ex->kind != irp::ExpKind::CONSTI && right->exp->u.ex->kind != irp::ExpKind::CONSTF)) {
                std::cerr << "AST2IRP::transCondExp(): right operand is not a constant" << std::endl;
                std::exit(1);
            }
            bool isRightTrue = right->exp->u.ex->kind == irp::ExpKind::CONSTI ? right->exp->u.ex->u.consti != 0
                                                                              : right->exp->u.ex->u.constf != 0.0;
            if (isRightTrue) {
                return new ExpTy(new env::Type(env::TyKind::INT, true), new TrExp(irp::ConstIExp(1)));
            } else {
                TrCx *leftCx = unCx(left->exp);
                return new ExpTy(new env::Type(env::TyKind::VOID, false),
                                 new TrExp(leftCx->trues, leftCx->falses, leftCx->stmt));
            }
        } else {
            TrCx *leftCx = unCx(left->exp);
            TrCx *rightCx = unCx(right->exp);

            temp::Label *rightLabel = temp::newLabel();
            doPatch(leftCx->falses, rightLabel);

            PatchList *trues = joinPatch(leftCx->trues, rightCx->trues);
            PatchList *falses = rightCx->falses;

            return new ExpTy(
                new env::Type(env::TyKind::INT, false),
                new TrExp(trues, falses,
                          irp::SeqStmt(leftCx->stmt, irp::SeqStmt(irp::LabelStmt(rightLabel), rightCx->stmt))));
        }
    }
    default:
        assert(0);
    }
}

ExpTy *AST2IRP::transIdExp(ast::Exp *exp) {
    if (!exp) {
        return nullptr;
    }

    // first, check if it is a local variable
    auto localVar = this->localTable.look(*exp->u.id);
    if (localVar) {
        if (localVar->type->isConst) {
            switch (localVar->type->type) {
            case env::TyKind::INT:
                return new ExpTy(localVar->type, new TrExp(irp::ConstIExp(localVar->u.consti)));
            case env::TyKind::FLOAT:
                return new ExpTy(localVar->type, new TrExp(irp::ConstFExp(localVar->u.constf)));
            case env::TyKind::INT_PTR:
            case env::TyKind::FLOAT_PTR:
                return new ExpTy(localVar->type, new TrExp(irp::NameExp(localVar->u.label)));
            default:
                assert(0);
            }
        }
        return new ExpTy(localVar->type, new TrExp(irp::TempExp(localVar->u.temp)));
    }

    // then, check if it is a global variable
    auto globalVar = this->globalTable.look(*exp->u.id);
    if (globalVar) {
        if (globalVar->type->isConst) {
            switch (globalVar->type->type) {
            case env::TyKind::INT:
                return new ExpTy(globalVar->type, new TrExp(irp::ConstIExp(globalVar->u.consti)));
            case env::TyKind::FLOAT:
                return new ExpTy(globalVar->type, new TrExp(irp::ConstFExp(globalVar->u.constf)));
            case env::TyKind::INT_PTR:
            case env::TyKind::FLOAT_PTR:
                return new ExpTy(globalVar->type, new TrExp(irp::NameExp(globalVar->u.label)));
            default:
                assert(0);
            }
        } else {
            switch (globalVar->type->type) {
            case env::TyKind::INT:
                return new ExpTy(globalVar->type,
                                 new TrExp(irp::MemExp(irp::NameExp(globalVar->u.label), irp::ExpType::INT)));
            case env::TyKind::FLOAT:
                return new ExpTy(globalVar->type,
                                 new TrExp(irp::MemExp(irp::NameExp(globalVar->u.label), irp::ExpType::FLOAT)));
            case env::TyKind::INT_PTR:
            case env::TyKind::FLOAT_PTR:
                return new ExpTy(globalVar->type, new TrExp(irp::NameExp(globalVar->u.label)));
            default:
                assert(0);
            }
        }
    }

    return nullptr;
}

ExpTy *AST2IRP::transIntConstExp(ast::Exp *exp) {
    if (!exp) {
        return nullptr;
    }
    return new ExpTy(new env::Type(env::TyKind::INT, true), new TrExp(irp::ConstIExp(exp->u.int_val)));
}

ExpTy *AST2IRP::transFloatConstExp(ast::Exp *exp) {
    if (!exp) {
        return nullptr;
    }
    return new ExpTy(new env::Type(env::TyKind::FLOAT, true),
                     new TrExp(irp::ConstFExp(std::stof(*(exp->u.float_val)))));
}

ExpTy *AST2IRP::transNotExp(ast::Exp *exp) {
    if (!exp) {
        return nullptr;
    }
    ExpTy *operand = transExp(exp->u.exp);
    if (!operand) {
        return nullptr;
    }
    irp::Exp *operandExp = unEx(operand->exp);

    if (operand->ty->isConst) {
        switch (operandExp->type) {
        case irp::ExpType::INT:
            return new ExpTy(new env::Type(env::TyKind::INT, true), new TrExp(irp::ConstIExp(!operandExp->u.consti)));
        case irp::ExpType::FLOAT:
            return new ExpTy(new env::Type(env::TyKind::INT, true), new TrExp(irp::ConstIExp(!operandExp->u.constf)));
        default:
            assert(0);
        }
    }

    TrCx *cx = unCx(operand->exp);
    return new ExpTy(new env::Type(env::TyKind::INT, false), new TrExp(cx->falses, cx->trues, cx->stmt));
}

ExpTy *AST2IRP::transMinusExp(ast::Exp *exp) {
    if (!exp) {
        return nullptr;
    }
    ExpTy *operand = transExp(exp->u.exp);
    if (!operand) {
        return nullptr;
    }
    irp::Exp *operandExp = unEx(operand->exp);

    if (operand->ty->isConst) {
        switch (operandExp->type) {
        case irp::ExpType::INT:
            return new ExpTy(new env::Type(env::TyKind::INT, true), new TrExp(irp::ConstIExp(-operandExp->u.consti)));
        case irp::ExpType::FLOAT:
            return new ExpTy(new env::Type(env::TyKind::FLOAT, true), new TrExp(irp::ConstFExp(-operandExp->u.constf)));
        default:
            assert(0);
        }
    }

    return new ExpTy(operand->ty, new TrExp(irp::BinOpExp(irp::BinOp::SUB, irp::ConstIExp(0), operandExp)));
}

ExpTy *AST2IRP::transCallExp(ast::Exp *exp) {
    if (!exp) {
        return nullptr;
    }
    auto funcInfo = this->funcTable.look(*exp->u.call.funcname);
    if (!funcInfo) {
        return nullptr;
    }

    TrExpList *args = transArgs(exp->u.call.args, funcInfo->paramTypes);
    irp::ExpList *irpArgs = unTrExpList(args);

    if (funcInfo->retType == env::FuncRetType::VOID) {
        return new ExpTy(new env::Type(env::TyKind::VOID, false),
                         new TrExp(irp::VoidCallStmt(funcInfo->name, irpArgs)));
    } else if (funcInfo->retType == env::FuncRetType::INT) {
        return new ExpTy(new env::Type(env::TyKind::INT, false),
                         new TrExp(irp::CallExp(funcInfo->name, irpArgs, irp::ExpType::INT)));
    } else if (funcInfo->retType == env::FuncRetType::FLOAT) {
        return new ExpTy(new env::Type(env::TyKind::FLOAT, false),
                         new TrExp(irp::CallExp(funcInfo->name, irpArgs, irp::ExpType::FLOAT)));
    }

    return nullptr;
}

ExpTy *AST2IRP::transArrayExp(ast::Exp *exp) {
    if (!exp) {
        return nullptr;
    }
    ExpTy *array = transExp(exp->u.array.array);
    if (!array) {
        return nullptr;
    }
    irp::Exp *arrayExp = unEx(array->exp);
    TrExpList *indices = transExpList(exp->u.array.index);
    if (!indices) {
        return nullptr;
    }
    irp::ExpList *irpIndices = unTrExpList(indices);

    std::vector<int> dims;
    auto globalVar = this->globalTable.look(*exp->u.array.array->u.id);
    if (globalVar) {
        dims = *globalVar->dims;
    } else {
        auto localVar = this->localTable.look(*exp->u.array.array->u.id);
        if (localVar) {
            dims = *localVar->dims;
        }
    }

    // get offset exp
    irp::Exp *offset = nullptr;
    if (dims.size() == 1) {
        offset = irp::BinOpExp(irp::BinOp::MUL, irpIndices->front(), irp::ConstIExp(4));
    } else {
        int step = std::accumulate(dims.begin(), dims.end(), 1, std::multiplies<int>());
        for (int i = 0; i < irpIndices->size(); ++i) {
            step /= dims[i];
            if (i == 0) {
                offset = irp::BinOpExp(irp::BinOp::MUL, (*irpIndices)[i], irp::ConstIExp(step));
            } else {
                offset = irp::BinOpExp(irp::BinOp::ADD, offset,
                                       irp::BinOpExp(irp::BinOp::MUL, (*irpIndices)[i], irp::ConstIExp(step)));
            }
        }
        offset = irp::BinOpExp(irp::BinOp::MUL, offset, irp::ConstIExp(4));
    }

    if (dims.size() == irpIndices->size()) {
        switch (array->ty->type) {
        case env::TyKind::INT_PTR: {
            return new ExpTy(
                new env::Type(env::TyKind::INT, false),
                new TrExp(irp::MemExp(irp::BinOpExp(irp::BinOp::ADD, arrayExp, offset), irp::ExpType::INT)));
        }
        case env::TyKind::FLOAT_PTR: {
            return new ExpTy(
                new env::Type(env::TyKind::FLOAT, false),
                new TrExp(irp::MemExp(irp::BinOpExp(irp::BinOp::ADD, arrayExp, offset), irp::ExpType::FLOAT)));
        }
        default:
            assert(0);
        }
    } else {
        switch (array->ty->type) {
        case env::TyKind::INT_PTR: {
            return new ExpTy(new env::Type(env::TyKind::INT_PTR, false),
                             new TrExp(irp::BinOpExp(irp::BinOp::ADD, arrayExp, offset)));
        }
        case env::TyKind::FLOAT_PTR: {
            return new ExpTy(new env::Type(env::TyKind::FLOAT_PTR, false),
                             new TrExp(irp::BinOpExp(irp::BinOp::ADD, arrayExp, offset)));
        }
        default:
            assert(0);
        }
    }

    return nullptr;
}

ExpTy *AST2IRP::transGetintExp() {
    return new ExpTy(new env::Type(env::TyKind::INT, false), new TrExp(irp::GetintExp()));
}

ExpTy *AST2IRP::transGetchExp() {
    return new ExpTy(new env::Type(env::TyKind::INT, false), new TrExp(irp::GetchExp()));
}

ExpTy *AST2IRP::transGetfloatExp() {
    return new ExpTy(new env::Type(env::TyKind::FLOAT, false), new TrExp(irp::GetfloatExp()));
}

ExpTy *AST2IRP::transGetarrayExp(ast::Exp *exp) {
    if (!exp) {
        return nullptr;
    }
    ExpTy *array = transExp(exp->u.exp);
    if (!array) {
        return nullptr;
    }
    return new ExpTy(new env::Type(env::TyKind::INT, false), new TrExp(irp::GetarrayExp(unEx(array->exp))));
}

ExpTy *AST2IRP::transGetfarrayExp(ast::Exp *exp) {
    if (!exp) {
        return nullptr;
    }
    ExpTy *array = transExp(exp->u.exp);
    if (!array) {
        return nullptr;
    }
    return new ExpTy(new env::Type(env::TyKind::FLOAT, false), new TrExp(irp::GetfarrayExp(unEx(array->exp))));
}

TrExpList *AST2IRP::transExpList(ast::ExpList *args) {
    if (!args) {
        return nullptr;
    }
    TrExpList *trExpList = new TrExpList();
    for (auto arg : args->exps) {
        ExpTy *exp = transExp(arg);
        if (!exp) {
            continue;
        }
        trExpList->push_back(exp->exp);
    }
    return trExpList;
}

TrExpList *AST2IRP::transArgs(ast::ExpList *args, std::vector<env::TyKind> *paramTypes) {
    if (!args) {
        return nullptr;
    }
    TrExpList *trExpList = new TrExpList();
    auto it = paramTypes->begin();
    for (auto arg : args->exps) {
        ExpTy *exp = transExp(arg);
        if (!exp) {
            continue;
        }
        if (exp->ty->type == env::TyKind::INT && *it == env::TyKind::FLOAT) {
            exp->exp = new TrExp(irp::CastExp(unEx(exp->exp), irp::ExpType::FLOAT));
        } else if (exp->ty->type == env::TyKind::FLOAT && *it == env::TyKind::INT) {
            exp->exp = new TrExp(irp::CastExp(unEx(exp->exp), irp::ExpType::INT));
        }
        trExpList->push_back(exp->exp);
        ++it;
    }
    return trExpList;
}

} // namespace irpgen