#include "canon.hpp"
#include "irp.hpp"
#include "temp.hpp"

#define CANON_DEBUG

namespace canon {
/* linearize */
struct StmExp {
    irp::Stmt *stm;
    irp::Exp *exp;
    StmExp(irp::Stmt *stm, irp::Exp *exp) : stm(stm), exp(exp) {}
};

struct ExpRefList {
    irp::Exp **head;
    ExpRefList *tail;
    ExpRefList(irp::Exp **head, ExpRefList *tail) : head(head), tail(tail) {}
};

static bool isNop(irp::Stmt *stmt) {
    return stmt->kind == irp::StmtKind::EXP &&
           (stmt->u.exp->kind == irp::ExpKind::CONSTI || stmt->u.exp->kind == irp::ExpKind::CONSTF);
}

static irp::Stmt *seq(irp::Stmt *s1, irp::Stmt *s2) {
    if (isNop(s1)) {
        return s2;
    }
    if (isNop(s2)) {
        return s1;
    }
    return irp::SeqStmt(s1, s2);
}

static bool commute(irp::Stmt *stmt, irp::Exp *exp) {
    if (isNop(stmt)) {
        return true;
    }
    if (exp->kind == irp::ExpKind::NAME || exp->kind == irp::ExpKind::CONSTI || exp->kind == irp::ExpKind::CONSTF) {
        return true;
    }
    return false;
}

static bool isCall(irp::Exp *exp) {
    return exp->kind == irp::ExpKind::CALL || exp->kind == irp::ExpKind::GETARRAY ||
           exp->kind == irp::ExpKind::GETFARRAY;
}

static StmExp do_exp(irp::Exp *exp);
static irp::Stmt *do_stm(irp::Stmt *stmt);

static irp::Stmt *reorder(ExpRefList *refList) {
#ifdef CANON_DEBUG
    std::cerr << "canon::reorder" << std::endl;
#endif
    if (!refList) {
        return irp::ExpStmt(irp::ConstIExp(0));
    }
    if (isCall(*refList->head)) {
#ifdef CANON_DEBUG
        std::cerr << "canon::reorder: isCall" << std::endl;
#endif
        temp::Temp *t =
            temp::newTemp((*refList->head)->type == irp::ExpType::INT ? temp::TempType::INT : temp::TempType::FLOAT);
        *refList->head = irp::EseqExp(irp::MoveStmt(irp::TempExp(t), *refList->head), irp::TempExp(t));
        return reorder(refList);
    } else {
        StmExp hd = do_exp(*refList->head);
        irp::Stmt *s = reorder(refList->tail);
        if (commute(s, hd.exp)) {
            *refList->head = hd.exp;
            return seq(hd.stm, s);
        } else {
            temp::Temp *t =
                temp::newTemp(hd.exp->type == irp::ExpType::INT ? temp::TempType::INT : temp::TempType::FLOAT);
            *refList->head = irp::TempExp(t);
            return seq(hd.stm, seq(irp::MoveStmt(irp::TempExp(t), hd.exp), s));
        }
    }
}

static ExpRefList *get_call_rlist(irp::ExpList *args) {
    ExpRefList *rlist, *curr;
    if (!args) {
        return nullptr;
    }

    curr = rlist = new ExpRefList(&args->front(), nullptr);
    for (auto it = args->begin() + 1; it != args->end(); ++it) {
        curr = curr->tail = new ExpRefList(&*it, nullptr);
    }
    return rlist;
}

static StmExp do_exp(irp::Exp *exp) {
    switch (exp->kind) {
    case irp::ExpKind::BINOP: {
        return StmExp(reorder(new ExpRefList(&exp->u.binop.left, new ExpRefList(&exp->u.binop.right, nullptr))), exp);
    }
    case irp::ExpKind::MEM: {
        return StmExp(reorder(new ExpRefList(&exp->u.mem, nullptr)), exp);
    }
    case irp::ExpKind::ESEQ: {
        StmExp x = do_exp(exp->u.eseq.exp);
        return StmExp(seq(do_stm(exp->u.eseq.stmt), x.stm), x.exp);
    }
    case irp::ExpKind::CAST: {
        return StmExp(reorder(new ExpRefList(&exp->u.cast, nullptr)), exp);
    }
    case irp::ExpKind::CALL: {
        return StmExp(reorder(get_call_rlist(exp->u.call.args)), exp);
    }
    case irp::ExpKind::GETARRAY: {
        return StmExp(reorder(new ExpRefList(&exp->u.getarray, nullptr)), exp);
    }
    case irp::ExpKind::GETFARRAY: {
        return StmExp(reorder(new ExpRefList(&exp->u.getarray, nullptr)), exp);
    }
    default: {
        return StmExp(reorder(nullptr), exp);
    }
    }
}

static irp::Stmt *do_stm(irp::Stmt *stmt) {
    switch (stmt->kind) {
    case irp::StmtKind::SEQ: {
#ifdef CANON_DEBUG
        std::cerr << "canon::do_stm: SEQ" << std::endl;
#endif
        return seq(do_stm(stmt->u.seq.left), do_stm(stmt->u.seq.right));
    }
    case irp::StmtKind::JUMP: {
#ifdef CANON_DEBUG
        std::cerr << "canon::do_stm: JUMP" << std::endl;
#endif
        return seq(reorder(nullptr), stmt);
    }
    case irp::StmtKind::CJUMP: {
#ifdef CANON_DEBUG
        std::cerr << "canon::do_stm: CJUMP" << std::endl;
#endif
        return seq(reorder(new ExpRefList(&stmt->u.cjump.left, new ExpRefList(&stmt->u.cjump.right, nullptr))), stmt);
    }
    case irp::StmtKind::EXP: {
#ifdef CANON_DEBUG
        std::cerr << "canon::do_stm: EXP" << std::endl;
#endif
        if (stmt->u.exp->kind == irp::ExpKind::CALL) {
            return seq(reorder(get_call_rlist(stmt->u.exp->u.call.args)), stmt);
        }
        if (stmt->u.exp->kind == irp::ExpKind::GETARRAY || stmt->u.exp->kind == irp::ExpKind::GETFARRAY) {
            return seq(reorder(new ExpRefList(&stmt->u.exp->u.getarray, nullptr)), stmt);
        }
        return seq(reorder(new ExpRefList(&stmt->u.exp, nullptr)), stmt);
    }
    case irp::StmtKind::RETURN: {
#ifdef CANON_DEBUG
        std::cerr << "canon::do_stm: RETURN" << std::endl;
#endif
        return seq(reorder(new ExpRefList(&stmt->u.exp, nullptr)), stmt);
    }
    case irp::StmtKind::VOID_CALL: {
        return seq(reorder(get_call_rlist(stmt->u.void_call.args)), stmt);
    }
    case irp::StmtKind::PUTINT: {
        return seq(reorder(new ExpRefList(&stmt->u.exp, nullptr)), stmt);
    }
    case irp::StmtKind::PUTCH: {
        return seq(reorder(new ExpRefList(&stmt->u.exp, nullptr)), stmt);
    }
    case irp::StmtKind::PUTFLOAT: {
        return seq(reorder(new ExpRefList(&stmt->u.exp, nullptr)), stmt);
    }
    case irp::StmtKind::PUTARRAY: {
        return seq(reorder(new ExpRefList(&stmt->u.putarray.outlen, new ExpRefList(&stmt->u.putarray.array, nullptr))),
                   stmt);
    }
    case irp::StmtKind::PUTFARRAY: {
        return seq(reorder(new ExpRefList(&stmt->u.putarray.outlen, new ExpRefList(&stmt->u.putarray.array, nullptr))),
                   stmt);
    }
    case irp::StmtKind::PUTF: {
        return seq(reorder(get_call_rlist(stmt->u.putf.args)), stmt);
    }
    case irp::StmtKind::MOVE: {
        if (stmt->u.move.dst->kind == irp::ExpKind::TEMP) {
#ifdef CANON_DEBUG
            std::cerr << "canon::do_stm: MOVE (dst=[TEMP," << stmt->u.move.dst->u.temp->num << "])" << std::endl;
#endif
            if (stmt->u.move.src->kind == irp::ExpKind::CALL) {
                return seq(reorder(get_call_rlist(stmt->u.move.src->u.call.args)), stmt);
            }
            if (stmt->u.move.src->kind == irp::ExpKind::GETARRAY || stmt->u.move.src->kind == irp::ExpKind::GETFARRAY) {
                return seq(reorder(new ExpRefList(&stmt->u.move.src->u.getarray, nullptr)), stmt);
            }
            return seq(reorder(new ExpRefList(&stmt->u.move.src, nullptr)), stmt);
        }
        if (stmt->u.move.dst->kind == irp::ExpKind::MEM) {
#ifdef CANON_DEBUG
            std::cerr << "canon::do_stm: MOVE: dst=MEM" << std::endl;
#endif
            return seq(reorder(new ExpRefList(&stmt->u.move.dst->u.mem, new ExpRefList(&stmt->u.move.src, nullptr))),
                       stmt);
        }
        if (stmt->u.move.dst->kind == irp::ExpKind::ESEQ) {
#ifdef CANON_DEBUG
            std::cerr << "canon::do_stm: MOVE: dst=ESEQ" << std::endl;
#endif
            irp::Stmt *s = stmt->u.move.dst->u.eseq.stmt;
            stmt->u.move.dst = stmt->u.move.dst->u.eseq.exp;
            return do_stm(irp::SeqStmt(s, stmt));
        }
        std::cerr << "do_stm: MOVE: unknown kind\n";
        exit(1);
    }
    default: {
        return stmt;
    }
    }
}

static irp::StmtList *linear(irp::Stmt *stmt, irp::StmtList *right) {
#ifdef CANON_DEBUG
    std::cerr << "canon::linear: start\n";
#endif
    if (stmt->kind == irp::StmtKind::SEQ) {
        return linear(stmt->u.seq.left, linear(stmt->u.seq.right, right));
    }
    return new irp::StmtList(stmt, right);
}

irp::StmtList *linearize(irp::Stmt *stmt) {
#ifdef CANON_DEBUG
    std::cerr << "canon::linearize: start\n";
#endif
    if (!stmt) {
        return nullptr;
    }
    return linear(do_stm(stmt), nullptr);
}

} // namespace canon