#include "irp.hpp"
#include <cassert>

namespace irp {
static bool isExpConst(Exp *exp) { return exp->kind == ExpKind::CONSTI || exp->kind == ExpKind::CONSTF; }

FuncDef *getFuncDef(std::string *name, temp::TempList *params, Stmt *body, FuncType type) {
    FuncDef *funcDef = new FuncDef();
    funcDef->name = name;
    funcDef->params = params;
    funcDef->body = body;
    funcDef->type = type;
    return funcDef;
}

GlobalDef *GlobalConstArrDef(temp::Label *label, std::string *llvmStr, std::string *armStr) {
    GlobalDef *globalDef = new GlobalDef();
    globalDef->kind = GlobalDefKind::CONST_ARR;
    globalDef->label = label;
    globalDef->llvmStr = llvmStr;
    globalDef->armStr = armStr;
    return globalDef;
}

GlobalDef *GlobalVarDef(temp::Label *label, std::string *llvmStr, std::string *armStr) {
    GlobalDef *globalDef = new GlobalDef();
    globalDef->kind = GlobalDefKind::VAR;
    globalDef->label = label;
    globalDef->llvmStr = llvmStr;
    globalDef->armStr = armStr;
    return globalDef;
}

GlobalDef *GlobalArrDef(temp::Label *label, std::string *llvmStr, std::string *armStr) {
    GlobalDef *globalDef = new GlobalDef();
    globalDef->kind = GlobalDefKind::ARR;
    globalDef->label = label;
    globalDef->llvmStr = llvmStr;
    globalDef->armStr = armStr;
    return globalDef;
}

LocalConstArrayDef *LocalConstArrDef(temp::Label *label, std::string *llvmStr, std::string *armStr) {
    LocalConstArrayDef *localConstArrayDef = new LocalConstArrayDef();
    localConstArrayDef->label = label;
    localConstArrayDef->llvmStr = llvmStr;
    localConstArrayDef->armStr = armStr;
    return localConstArrayDef;
}

Stmt *SeqStmt(Stmt *left, Stmt *right) {
    Stmt *stmt = new Stmt();
    stmt->kind = StmtKind::SEQ;
    stmt->u.seq.left = left;
    stmt->u.seq.right = right;
    return stmt;
}

Stmt *LabelStmt(temp::Label *label) {
    Stmt *stmt = new Stmt();
    stmt->kind = StmtKind::LABEL;
    stmt->u.label = label;
    return stmt;
}

Stmt *MoveStmt(Exp *dst, Exp *src) {
    Stmt *stmt = new Stmt();
    stmt->kind = StmtKind::MOVE;
    stmt->u.move.dst = dst;
    stmt->u.move.src = src;
    return stmt;
}

Stmt *ExpStmt(Exp *exp) {
    Stmt *stmt = new Stmt();
    stmt->kind = StmtKind::EXP;
    stmt->u.exp = exp;
    return stmt;
}

Stmt *ReturnStmt(Exp *exp) {
    Stmt *stmt = new Stmt();
    stmt->kind = StmtKind::RETURN;
    stmt->u.exp = exp;
    return stmt;
}

Stmt *JumpStmt(temp::Label *jump) {
    Stmt *stmt = new Stmt();
    stmt->kind = StmtKind::JUMP;
    stmt->u.jump = jump;
    return stmt;
}

Stmt *CJumpStmt(RelOp op, Exp *left, Exp *right, temp::Label *t, temp::Label *f) {
    if (left->type == ExpType::FLOAT || right->type == ExpType::FLOAT) {
        left = CastExp(left, ExpType::FLOAT);
        right = CastExp(right, ExpType::FLOAT);
    }

    // if (isExpConst(left) && isExpConst(right)) {
    //     bool result = false;
    //     switch (op) {
    //     case RelOp::EQ:
    //         result =
    //             left->kind == ExpKind::CONSTI ? left->u.consti == right->u.consti : left->u.constf == right->u.constf;
    //         break;
    //     case RelOp::NE:
    //         result =
    //             left->kind == ExpKind::CONSTI ? left->u.consti != right->u.consti : left->u.constf != right->u.constf;
    //         break;
    //     case RelOp::LT:
    //         result =
    //             left->kind == ExpKind::CONSTI ? left->u.consti < right->u.consti : left->u.constf < right->u.constf;
    //         break;
    //     case RelOp::GT:
    //         result =
    //             left->kind == ExpKind::CONSTI ? left->u.consti > right->u.consti : left->u.constf > right->u.constf;
    //         break;
    //     case RelOp::LE:
    //         result =
    //             left->kind == ExpKind::CONSTI ? left->u.consti <= right->u.consti : left->u.constf <= right->u.constf;
    //         break;
    //     case RelOp::GE:
    //         result =
    //             left->kind == ExpKind::CONSTI ? left->u.consti >= right->u.consti : left->u.constf >= right->u.constf;
    //         break;
    //     default:
    //         assert(0);
    //     }
    //     return result ? JumpStmt(t) : JumpStmt(f);
    // }

    Stmt *stmt = new Stmt();
    stmt->kind = StmtKind::CJUMP;
    stmt->u.cjump.op = op;
    stmt->u.cjump.left = left;
    stmt->u.cjump.right = right;
    stmt->u.cjump.t = t;
    stmt->u.cjump.f = f;
    return stmt;
}

Stmt *VoidCallStmt(std::string *name, ExpList *args) {
    Stmt *stmt = new Stmt();
    stmt->kind = StmtKind::VOID_CALL;
    stmt->u.void_call.name = name;
    stmt->u.void_call.args = args;
    return stmt;
}

Stmt *PutintStmt(Exp *exp) {
    Stmt *stmt = new Stmt();
    stmt->kind = StmtKind::PUTINT;
    stmt->u.exp = exp;
    return stmt;
}

Stmt *PutchStmt(Exp *exp) {
    Stmt *stmt = new Stmt();
    stmt->kind = StmtKind::PUTCH;
    stmt->u.exp = exp;
    return stmt;
}

Stmt *PutfloatStmt(Exp *exp) {
    Stmt *stmt = new Stmt();
    stmt->kind = StmtKind::PUTFLOAT;
    stmt->u.exp = exp;
    return stmt;
}

Stmt *PutarrayStmt(Exp *array, Exp *outlen) {
    Stmt *stmt = new Stmt();
    stmt->kind = StmtKind::PUTARRAY;
    stmt->u.putarray.array = array;
    stmt->u.putarray.outlen = outlen;
    return stmt;
}

Stmt *PutfarrayStmt(Exp *array, Exp *outlen) {
    Stmt *stmt = new Stmt();
    stmt->kind = StmtKind::PUTFARRAY;
    stmt->u.putarray.array = array;
    stmt->u.putarray.outlen = outlen;
    return stmt;
}

Stmt *PutfStmt(std::string *fmt, ExpList *args) {
    Stmt *stmt = new Stmt();
    stmt->kind = StmtKind::PUTF;
    stmt->u.putf.fmt = fmt;
    stmt->u.putf.args = args;
    return stmt;
}

Stmt *StarttimeStmt() {
    Stmt *stmt = new Stmt();
    stmt->kind = StmtKind::STARTTIME;
    return stmt;
}

Stmt *StoptimeStmt() {
    Stmt *stmt = new Stmt();
    stmt->kind = StmtKind::STOPTIME;
    return stmt;
}

Stmt *MemsetStmt(temp::Temp *array, int offset, int size) {
    Stmt *stmt = new Stmt();
    stmt->kind = StmtKind::MEMSET;
    stmt->u.memset.array = array;
    stmt->u.memset.offset = offset;
    stmt->u.memset.size = size;
    return stmt;
}

Exp *BinOpExp(BinOp op, Exp *left, Exp *right) {
    if (isExpConst(left) && isExpConst(right)) {
        if (left->kind == ExpKind::CONSTI && right->kind == ExpKind::CONSTI) {
            switch (op) {
            case BinOp::ADD:
                return ConstIExp(left->u.consti + right->u.consti);
            case BinOp::SUB:
                return ConstIExp(left->u.consti - right->u.consti);
            case BinOp::MUL:
                return ConstIExp(left->u.consti * right->u.consti);
            case BinOp::DIV:
                return ConstIExp(left->u.consti / right->u.consti);
            case BinOp::MOD:
                return ConstIExp(left->u.consti % right->u.consti);
            default:
                assert(0);
            }
        } else {
            float l = left->kind == ExpKind::CONSTI ? static_cast<float>(left->u.consti) : left->u.constf;
            float r = right->kind == ExpKind::CONSTI ? static_cast<float>(right->u.consti) : right->u.constf;
            switch (op) {
            case BinOp::ADD:
                return ConstFExp(l + r);
            case BinOp::SUB:
                return ConstFExp(l - r);
            case BinOp::MUL:
                return ConstFExp(l * r);
            case BinOp::DIV:
                return ConstFExp(l / r);
            default:
                assert(0);
            }
        }
    } else if (isExpConst(left)) {
        if (left->kind == ExpKind::CONSTI) {
            if (left->u.consti == 0) {
                if (op == BinOp::ADD) {
                    return right;
                } else if (op == BinOp::MUL) {
                    return ConstIExp(0);
                } else if (op == BinOp::DIV) {
                    return ConstIExp(0);
                } else if (op == BinOp::MOD) {
                    return ConstIExp(0);
                }
            } else if (left->u.consti == 1) {
                if (op == BinOp::MUL) {
                    return right;
                } else if (op == BinOp::DIV) {
                    return right;
                }
            }
        }
    } else if (isExpConst(right)) {
        if (right->kind == ExpKind::CONSTI) {
            if (right->u.consti == 0) {
                if (op == BinOp::ADD) {
                    return left;
                } else if (op == BinOp::SUB) {
                    return left;
                } else if (op == BinOp::MUL) {
                    return ConstIExp(0);
                }
            } else if (right->u.consti == 1) {
                if (op == BinOp::MUL) {
                    return left;
                }
            }
        }
    }

    Exp *exp = new Exp();
    exp->kind = ExpKind::BINOP;
    exp->type = ExpType::INT;
    if (op != BinOp::DIV && (left->type == ExpType::FLOAT || right->type == ExpType::FLOAT)) {
        left = CastExp(left, ExpType::FLOAT);
        right = CastExp(right, ExpType::FLOAT);
        exp->type = ExpType::FLOAT;
    }
    exp->u.binop.op = op;
    exp->u.binop.left = left;
    exp->u.binop.right = right;
    return exp;
}

Exp *MemExp(Exp *mem, ExpType type) {
    Exp *exp = new Exp();
    exp->kind = ExpKind::MEM;
    exp->type = type;
    exp->u.mem = mem;
    return exp;
}

Exp *TempExp(temp::Temp *temp) {
    Exp *exp = new Exp();
    exp->kind = ExpKind::TEMP;
    exp->type = temp->type == temp::TempType::FLOAT ? ExpType::FLOAT : ExpType::INT;
    exp->u.temp = temp;
    return exp;
}

Exp *EseqExp(Stmt *stmt, Exp *exp) {
    Exp *e = new Exp();
    e->kind = ExpKind::ESEQ;
    e->type = exp->type;
    e->u.eseq.stmt = stmt;
    e->u.eseq.exp = exp;
    return e;
}

Exp *NameExp(temp::Label *name) {
    Exp *exp = new Exp();
    exp->kind = ExpKind::NAME;
    exp->type = ExpType::INT;
    exp->u.name = name;
    return exp;
}

Exp *ConstIExp(int consti) {
    Exp *exp = new Exp();
    exp->kind = ExpKind::CONSTI;
    exp->type = ExpType::INT;
    exp->u.consti = consti;
    return exp;
}

Exp *ConstFExp(float constf) {
    Exp *exp = new Exp();
    exp->kind = ExpKind::CONSTF;
    exp->type = ExpType::FLOAT;
    exp->u.constf = constf;
    return exp;
}

Exp *CallExp(std::string *name, ExpList *args, ExpType type) {
    Exp *exp = new Exp();
    exp->kind = ExpKind::CALL;
    exp->type = type;
    exp->u.call.name = name;
    exp->u.call.args = args;
    return exp;
}

Exp *CastExp(Exp *exp, ExpType type) {
    if (exp->type == type) {
        return exp;
    }
    if (exp->kind == ExpKind::CONSTI) {
        return ConstFExp(static_cast<float>(exp->u.consti));
    }
    if (exp->kind == ExpKind::CONSTF) {
        return ConstIExp(static_cast<int>(exp->u.constf));
    }
    Exp *e = new Exp();
    e->kind = ExpKind::CAST;
    e->type = type;
    e->u.cast = exp;
    return e;
}

Exp *GetintExp() {
    Exp *exp = new Exp();
    exp->kind = ExpKind::GETINT;
    exp->type = ExpType::INT;
    return exp;
}

Exp *GetchExp() {
    Exp *exp = new Exp();
    exp->kind = ExpKind::GETCH;
    exp->type = ExpType::INT;
    return exp;
}

Exp *GetfloatExp() {
    Exp *exp = new Exp();
    exp->kind = ExpKind::GETFLOAT;
    exp->type = ExpType::FLOAT;
    return exp;
}

Exp *GetarrayExp(Exp *array) {
    Exp *exp = new Exp();
    exp->kind = ExpKind::GETARRAY;
    exp->type = ExpType::INT;
    exp->u.getarray = array;
    return exp;
}

Exp *GetfarrayExp(Exp *array) {
    Exp *exp = new Exp();
    exp->kind = ExpKind::GETFARRAY;
    exp->type = ExpType::INT;
    exp->u.getarray = array;
    return exp;
}

Exp *AllocaExp(int size, int offset) {
    Exp *exp = new Exp();
    exp->kind = ExpKind::ALLOCA;
    exp->type = ExpType::INT;
    exp->u.alloca.size = size;
    exp->u.alloca.frameOffset = offset;
    return exp;
}

} // namespace irp