#include "astPrinter.hpp"

namespace ast {
void Printer::print() { this->printCompUnitList(this->root->comp_unit_list); }

void Printer::printCompUnitList(CompUnitList *node) {
    if (!node) {
        return;
    }
    for (auto &comp_unit : node->comp_units) {
        this->printCompUnit(comp_unit);
    }
}

void Printer::printCompUnit(CompUnit *node) {
    if (!node) {
        return;
    }
    switch (node->kind) {
    case CompUnitKind::DECL:
        this->printDecl(node->u.decl);
        break;
    case CompUnitKind::FUNC_DEF:
        this->printFuncDef(node->u.func_def);
        break;
    }
    this->os << std::endl;
}

void Printer::printDecl(Decl *node) {
    if (!node) {
        return;
    }
    switch (node->kind) {
    case DeclKind::CONST:
        this->printConstDecl(node->u.const_decl);
        break;
    case DeclKind::VAR:
        this->printVarDecl(node->u.var_decl);
        break;
    }
}

void Printer::printConstDecl(ConstDecl *node) {
    if (!node) {
        return;
    }
    this->os << "const ";
    switch (node->type) {
    case BaseType::INT:
        this->os << "int ";
        break;
    case BaseType::FLOAT:
        this->os << "float ";
        break;
    }
    this->printConstDefList(node->const_defs);
    this->os << ";" << std::endl;
}

void Printer::printConstDefList(ConstDefList *node) {
    if (!node) {
        return;
    }
    // separated by commas
    for (auto it = node->const_defs.begin(); it != node->const_defs.end(); ++it) {
        this->printConstDef(*it);
        if (std::next(it) != node->const_defs.end()) {
            this->os << ", ";
        }
    }
}

void Printer::printConstDef(ConstDef *node) {
    if (!node) {
        return;
    }
    this->os << *(node->id);
    this->printArrayDim(node->array_dim);
    this->os << " = ";
    this->printInitVal(node->init_val);
}

void Printer::printArrayDim(ExpList *node) {
    if (!node) {
        return;
    }
    for (auto &exp : node->exps) {
        this->os << "[";
        this->printExp(exp);
        this->os << "]";
    }
}

void Printer::printVarDecl(VarDecl *node) {
    if (!node) {
        return;
    }
    switch (node->type) {
    case BaseType::INT:
        this->os << "int ";
        break;
    case BaseType::FLOAT:
        this->os << "float ";
        break;
    }
    this->printVarDefList(node->var_defs);
    this->os << ";" << std::endl;
}

void Printer::printVarDefList(VarDefList *node) {
    if (!node) {
        return;
    }
    // separated by commas
    for (auto it = node->var_defs.begin(); it != node->var_defs.end(); ++it) {
        this->printVarDef(*it);
        if (std::next(it) != node->var_defs.end()) {
            this->os << ", ";
        }
    }
}

void Printer::printVarDef(VarDef *node) {
    if (!node) {
        return;
    }
    this->os << *(node->id);
    this->printArrayDim(node->array_dim);
    if (node->init_val) {
        this->os << " = ";
        this->printInitVal(node->init_val);
    }
}

void Printer::printInitVal(InitVal *node) {
    if (!node) {
        return;
    }
    switch (node->kind) {
    case InitValKind::VAL:
        this->printExp(node->u.val);
        break;
    case InitValKind::ARRAY:
        this->os << "{";
        this->printInitValList(node->u.array);
        this->os << "}";
        break;
    }
}

void Printer::printInitValList(InitValList *node) {
    if (!node) {
        return;
    }
    // separated by commas
    for (auto it = node->array.begin(); it != node->array.end(); ++it) {
        this->printInitVal(*it);
        if (std::next(it) != node->array.end()) {
            this->os << ", ";
        }
    }
}

void Printer::printFuncDef(FuncDef *node) {
    if (!node) {
        return;
    }
    switch (node->ret_type) {
    case FuncType::INT:
        this->os << "int ";
        break;
    case FuncType::FLOAT:
        this->os << "float ";
        break;
    case FuncType::VOID:
        this->os << "void ";
        break;
    }
    this->os << *(node->id) << "(";
    this->printFuncFParamList(node->f_param_list);
    this->os << ") ";
    this->printBlock(node->block);
}

void Printer::printFuncFParamList(FuncFParamList *node) {
    if (!node) {
        return;
    }
    // separated by commas
    for (auto it = node->f_params.begin(); it != node->f_params.end(); ++it) {
        this->printFuncFParam(*it);
        if (std::next(it) != node->f_params.end()) {
            this->os << ", ";
        }
    }
}

void Printer::printFuncFParam(FuncFParam *node) {
    if (!node) {
        return;
    }
    switch (node->type) {
    case BaseType::INT:
        this->os << "int ";
        break;
    case BaseType::FLOAT:
        this->os << "float ";
        break;
    }
    this->os << *(node->id);
    if (node->is_array) {
        this->os << "[]";
        this->printArrayDim(node->array_dim);
    }
}

void Printer::printBlock(Block *node) {
    if (!node) {
        return;
    }
    this->os << "{" << std::endl;
    this->indent++;
    this->printBlockItemList(node->block_item_list);
    this->indent--;
    this->os << std::string(this->indent, '\t');
    this->os << "}" << std::endl;
}

void Printer::printBlockItemList(BlockItemList *node) {
    if (!node) {
        return;
    }
    for (auto &block_item : node->block_items) {
        this->printBlockItem(block_item);
    }
}

void Printer::printBlockItem(BlockItem *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t');
    switch (node->kind) {
    case BlockItemKind::DECL:
        this->printDecl(node->u.decl);
        break;
    case BlockItemKind::STMT:
        this->printStmt(node->u.stmt);
        break;
    }
}

void Printer::printStmt(Stmt *node) {
    if (!node) {
        return;
    }
    switch (node->kind) {
    case StmtKind::ASSIGN: {
        this->printExp(node->u.assign.lhs);
        this->os << " = ";
        this->printExp(node->u.assign.rhs);
        this->os << ";" << std::endl;
        break;
    }
    case StmtKind::EXP: {
        this->printExp(node->u.exp);
        this->os << ";" << std::endl;
        break;
    }
    case StmtKind::BLOCK:
        this->printBlock(node->u.block);
        break;
    case StmtKind::IF: {
        this->os << "if (";
        this->printExp(node->u.if_stmt.cond);
        this->os << ") ";
        this->printStmt(node->u.if_stmt.thenStmt);
        if (node->u.if_stmt.elseStmt) {
            this->os << "else ";
            this->printStmt(node->u.if_stmt.elseStmt);
        }
        break;
    }
    case StmtKind::WHILE: {
        this->os << "while (";
        this->printExp(node->u.while_stmt.cond);
        this->os << ") ";
        this->printStmt(node->u.while_stmt.body);
        break;
    }
    case StmtKind::BREAK:
        this->os << "break;" << std::endl;
        break;
    case StmtKind::CONTINUE:
        this->os << "continue;" << std::endl;
        break;
    case StmtKind::RETURN: {
        this->os << "return";
        if (node->u.exp) {
            this->os << " ";
            this->printExp(node->u.exp);
        }
        this->os << ";" << std::endl;
        break;
    }
    case StmtKind::PUTINT: {
        this->os << "putint(";
        this->printExp(node->u.exp);
        this->os << ");" << std::endl;
        break;
    }
    case StmtKind::PUTCH: {
        this->os << "putch(";
        this->printExp(node->u.exp);
        this->os << ");" << std::endl;
        break;
    }
    case StmtKind::PUTFLOAT: {
        this->os << "putfloat(";
        this->printExp(node->u.exp);
        this->os << ");" << std::endl;
        break;
    }
    case StmtKind::PUTARRAY: {
        this->os << "putarray(";
        this->printExp(node->u.putarray.out_len);
        this->os << ", ";
        this->printExp(node->u.putarray.array);
        this->os << ");" << std::endl;
        break;
    }
    case StmtKind::PUTFARRAY: {
        this->os << "putfarray(";
        this->printExp(node->u.putarray.out_len);
        this->os << ", ";
        this->printExp(node->u.putarray.array);
        this->os << ");" << std::endl;
        break;
    }
    case StmtKind::PUTF: {
        this->os << "putf(\"" << *(node->u.putf.fmt) << "\"";
        if (node->u.putf.args) {
            this->os << ", ";
            this->printArgs(node->u.putf.args);
        }
        this->os << ");" << std::endl;
        break;
    }
    case StmtKind::STARTTIME:
        this->os << "starttime();" << std::endl;
        break;
    case StmtKind::STOPTIME:
        this->os << "stoptime();" << std::endl;
        break;
    }
}

void Printer::printArgs(ExpList *node) {
    if (!node) {
        return;
    }
    // separated by commas
    for (auto it = node->exps.begin(); it != node->exps.end(); ++it) {
        this->printExp(*it);
        if (std::next(it) != node->exps.end()) {
            this->os << ", ";
        }
    }
}

void Printer::printExp(Exp *node) {
    if (!node) {
        return;
    }
    switch (node->kind) {
    case ExpKind::BINOP: {
        this->os << "(";
        this->printExp(node->u.binop.lhs);
        this->os << " ";
        this->printBinOp(node->u.binop.op);
        this->os << " ";
        this->printExp(node->u.binop.rhs);
        this->os << ")";
        break;
    }
    case ExpKind::COND: {
        this->os << "(";
        this->printExp(node->u.cond.lhs);
        this->os << " ";
        this->printRelOp(node->u.cond.op);
        this->os << " ";
        this->printExp(node->u.cond.rhs);
        this->os << ")";
        break;
    }
    case ExpKind::ID:
        this->os << *(node->u.id);
        break;
    case ExpKind::INTCONST:
        this->os << node->u.int_val;
        break;
    case ExpKind::FLOATCONST:
        this->os << *(node->u.float_val);
        break;
    case ExpKind::NOT: {
        this->os << "!";
        this->printExp(node->u.exp);
        break;
    }
    case ExpKind::MINUS: {
        this->os << "-";
        this->printExp(node->u.exp);
        break;
    }
    case ExpKind::CALL: {
        this->os << *(node->u.call.funcname) << "(";
        this->printArgs(node->u.call.args);
        this->os << ")";
        break;
    }
    case ExpKind::ARRAY: {
        this->printExp(node->u.array.array);
        this->printArrayDim(node->u.array.index);
        break;
    }
    case ExpKind::GETINT:
        this->os << "getint()";
        break;
    case ExpKind::GETCH:
        this->os << "getch()";
        break;
    case ExpKind::GETFLOAT:
        this->os << "getfloat()";
        break;
    case ExpKind::GETARRAY: {
        this->os << "getarray(";
        this->printExp(node->u.exp);
        this->os << ")";
        break;
    }
    case ExpKind::GETFARRAY: {
        this->os << "getfarray(";
        this->printExp(node->u.exp);
        this->os << ")";
        break;
    }
    }
}

void Printer::printBinOp(BinOp op) {
    switch (op) {
    case BinOp::ADD:
        this->os << "+";
        break;
    case BinOp::SUB:
        this->os << "-";
        break;
    case BinOp::MUL:
        this->os << "*";
        break;
    case BinOp::DIV:
        this->os << "/";
        break;
    case BinOp::MOD:
        this->os << "%";
        break;
    }
}

void Printer::printRelOp(RelOp op) {
    switch (op) {
    case RelOp::GT:
        this->os << ">";
        break;
    case RelOp::LT:
        this->os << "<";
        break;
    case RelOp::GE:
        this->os << ">=";
        break;
    case RelOp::LE:
        this->os << "<=";
        break;
    case RelOp::EQ:
        this->os << "==";
        break;
    case RelOp::NE:
        this->os << "!=";
        break;
    case RelOp::AND:
        this->os << "&&";
        break;
    case RelOp::OR:
        this->os << "||";
        break;
    }
}
} // namespace ast