#include "irpPrinter.hpp"
#include <iostream>

#define IRP_PRINTER_DEBUG
#undef IRP_PRINTER_DEBUG

namespace irp {
static std::string getTempTypeStr(temp::TempType type) {
    switch (type) {
    case temp::TempType::INT:
        return "INT";
    case temp::TempType::FLOAT:
        return "FLOAT";
    case temp::TempType::INT_PTR:
        return "INT_PTR";
    case temp::TempType::FLOAT_PTR:
        return "FLOAT_PTR";
    default:
        return "UNKNOWN";
    }
}

static std::string getExpTypeStr(ExpType type) {
    switch (type) {
    case ExpType::INT:
        return "INT";
    case ExpType::FLOAT:
        return "FLOAT";
    default:
        return "UNKNOWN";
    }
}

void Printer::print() {
    this->printGlobalDefs(root->globalDefs);
    this->printFuncDefList(root->funcs);
    this->printLocalConstDefs(root->localConstArrayDefs);
}

void Printer::printGlobalDefs(const std::vector<GlobalDef *> &globalTable) {
    this->os << "==================== Global Defs ====================" << std::endl;
    for (auto &p : globalTable) {
        this->os << *p->armStr << std::endl;
    }
}

void Printer::printLocalConstDefs(const std::vector<LocalConstArrayDef *> &localConstArrayTable) {
    this->os << "==================== Local Const Array Defs ====================" << std::endl;
    for (auto &p : localConstArrayTable) {
        this->os << *p->armStr << std::endl;
    }
}

void Printer::printFuncDefList(FuncDefList *node) {
    this->os << "==================== Func Defs ====================" << std::endl;
    if (!node) {
        return;
    }
    for (auto &func : *node) {
        printFuncDef(func);
    }
    this->os << std::endl;
}

void Printer::printFuncDef(FuncDef *node) {
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Printing FuncDef for " << *node->name << std::endl;
#endif
    if (!node) {
        return;
    }
    this->os << "<FuncDef>\n";

    this->indent++;
    this->os << std::string(this->indent, '\t') << "<name>" << *node->name << "</name>\n";
    this->printTempList(node->params);
    this->printStmt(node->body);
    this->indent--;
}

void Printer::printTempList(temp::TempList *node) {
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Printing TempList, with size=" << (node == nullptr ? 0 : node->size()) << std::endl;
#endif
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<params>\n";
    this->indent++;
    for (auto &temp : *node) {
        this->printTemp(temp);
    }
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</params>\n";
}

void Printer::printTemp(temp::Temp *node) {
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Printing Temp" << std::endl;
#endif
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<temp><name>" << node->num << "</name><type>"
             << getTempTypeStr(node->type) << "</type></temp>\n";
}

void Printer::printStmt(Stmt *node) {
    if (!node) {
        return;
    }
    switch (node->kind) {
    case StmtKind::SEQ:
        this->printSeqStmt(node);
        break;
    case StmtKind::LABEL:
        this->printLabelStmt(node);
        break;
    case StmtKind::MOVE:
        this->printMoveStmt(node);
        break;
    case StmtKind::EXP:
        this->printExpStmt(node);
        break;
    case StmtKind::RETURN:
        this->printReturnStmt(node);
        break;
    case StmtKind::JUMP:
        this->printJumpStmt(node);
        break;
    case StmtKind::CJUMP:
        this->printCJumpStmt(node);
        break;
    case StmtKind::VOID_CALL:
        this->printVoidCallStmt(node);
        break;
    case StmtKind::PUTINT:
        this->printPutintStmt(node);
        break;
    case StmtKind::PUTCH:
        this->printPutchStmt(node);
        break;
    case StmtKind::PUTFLOAT:
        this->printPutfloatStmt(node);
        break;
    case StmtKind::PUTARRAY:
        this->printPutarrayStmt(node);
        break;
    case StmtKind::PUTFARRAY:
        this->printPutfarrayStmt(node);
        break;
    case StmtKind::PUTF:
        this->printPutfStmt(node);
        break;
    case StmtKind::STARTTIME:
        this->printStarttimeStmt();
        break;
    case StmtKind::STOPTIME:
        this->printStoptimeStmt();
        break;
    case StmtKind::MEMSET:
        this->printMemsetStmt(node);
        break;
    default:
        break;
    }
}

void Printer::printSeqStmt(Stmt *node) {
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Printing SeqStmt" << std::endl;
#endif
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<SeqStmt>\n";
    this->indent++;
    this->printStmt(node->u.seq.left);
    this->printStmt(node->u.seq.right);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</SeqStmt>\n";
}

void Printer::printLabelStmt(Stmt *node) {
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Printing LabelStmt" << std::endl;
#endif
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Label>" << node->u.label->name << "</Label>\n";
}

void Printer::printMoveStmt(Stmt *node) {
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Printing MoveStmt" << std::endl;
#endif
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Move>\n";
    this->indent++;
    this->printExp(node->u.move.dst);
    this->printExp(node->u.move.src);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Move>\n";
}

void Printer::printExpStmt(Stmt *node) {
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Printing ExpStmt" << std::endl;
#endif
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Exp>\n";
    this->indent++;
    this->printExp(node->u.exp);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Exp>\n";
}

void Printer::printReturnStmt(Stmt *node) {
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Printing ReturnStmt" << std::endl;
#endif
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Return>\n";
    this->indent++;
    this->printExp(node->u.exp);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Return>\n";
}

void Printer::printJumpStmt(Stmt *node) {
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Printing JumpStmt" << std::endl;
#endif
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Jump>\n";
    this->indent++;
    this->os << std::string(this->indent, '\t') << "<Label>" << node->u.jump->name << "</Label>\n";
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Jump>\n";
}

void Printer::printCJumpStmt(Stmt *node) {
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Printing CJumpStmt" << std::endl;
#endif
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<CJump>\n";
    this->indent++;
    switch (node->u.cjump.op) {
    case RelOp::EQ:
        this->os << std::string(this->indent, '\t') << "<RelOp>EQ</RelOp>\n";
        break;
    case RelOp::NE:
        this->os << std::string(this->indent, '\t') << "<RelOp>NE</RelOp>\n";
        break;
    case RelOp::LT:
        this->os << std::string(this->indent, '\t') << "<RelOp>LT</RelOp>\n";
        break;
    case RelOp::GT:
        this->os << std::string(this->indent, '\t') << "<RelOp>GT</RelOp>\n";
        break;
    case RelOp::LE:
        this->os << std::string(this->indent, '\t') << "<RelOp>LE</RelOp>\n";
        break;
    case RelOp::GE:
        this->os << std::string(this->indent, '\t') << "<RelOp>GE</RelOp>\n";
        break;
    default:
        break;
    }
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Printing CJumpStmt left" << std::endl;
#endif
    this->printExp(node->u.cjump.left);
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Printing CJumpStmt right" << std::endl;
#endif
    this->printExp(node->u.cjump.right);
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Printing CJumpStmt true label" << (node->u.cjump.t == nullptr) << std::endl;
#endif
    this->os << std::string(this->indent, '\t') << "<TrueLabel>" << node->u.cjump.t->name << "</TrueLabel>\n";
    this->os << std::string(this->indent, '\t') << "<FalseLabel>" << node->u.cjump.f->name << "</FalseLabel>\n";
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</CJump>\n";
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Finished printing CJumpStmt" << std::endl;
#endif
}

void Printer::printVoidCallStmt(Stmt *node) {
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Printing VoidCallStmt" << std::endl;
#endif
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<VoidCall>\n";
    this->indent++;
    this->os << std::string(this->indent, '\t') << "<name>" << *node->u.void_call.name << "</name>\n";
    this->printExpList(node->u.void_call.args);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</VoidCall>\n";
}

void Printer::printPutintStmt(Stmt *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Putint>\n";
    this->indent++;
    this->printExp(node->u.exp);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Putint>\n";
}

void Printer::printPutchStmt(Stmt *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Putch>\n";
    this->indent++;
    this->printExp(node->u.exp);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Putch>\n";
}

void Printer::printPutfloatStmt(Stmt *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Putfloat>\n";
    this->indent++;
    this->printExp(node->u.exp);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Putfloat>\n";
}

void Printer::printPutarrayStmt(Stmt *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Putarray>\n";
    this->indent++;
    this->printExp(node->u.putarray.outlen);
    this->printExp(node->u.putarray.array);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Putarray>\n";
}

void Printer::printPutfarrayStmt(Stmt *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Putfarray>\n";
    this->indent++;
    this->printExp(node->u.putarray.outlen);
    this->printExp(node->u.putarray.array);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Putfarray>\n";
}

void Printer::printPutfStmt(Stmt *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Putf>\n";
    this->indent++;
    this->os << std::string(this->indent, '\t') << "<fmt>" << *node->u.putf.fmt << "</fmt>\n";
    this->printExpList(node->u.putf.args);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Putf>\n";
}

void Printer::printStarttimeStmt() { this->os << std::string(this->indent, '\t') << "<Starttime/>\n"; }

void Printer::printStoptimeStmt() { this->os << std::string(this->indent, '\t') << "<Stoptime/>\n"; }

void Printer::printMemsetStmt(Stmt *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Memset>\n";
    this->indent++;
    this->printTemp(node->u.memset.array);
    this->os << std::string(this->indent, '\t') << "<offset>" << node->u.memset.offset << "</offset>\n";
    this->os << std::string(this->indent, '\t') << "<size>" << node->u.memset.size * 4 << "</size>\n";
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Memset>\n";
}

void Printer::printExp(Exp *node) {
    if (!node) {
        return;
    }
    switch (node->kind) {
    case ExpKind::BINOP:
        this->printBinOpExp(node);
        break;
    case ExpKind::MEM:
        this->printMemExp(node);
        break;
    case ExpKind::TEMP:
        this->printTempExp(node);
        break;
    case ExpKind::ESEQ:
        this->printEseqExp(node);
        break;
    case ExpKind::NAME:
        this->printNameExp(node);
        break;
    case ExpKind::CONSTI:
        this->printConstIExp(node);
        break;
    case ExpKind::CONSTF:
        this->printConstFExp(node);
        break;
    case ExpKind::CALL:
        this->printCallExp(node);
        break;
    case ExpKind::CAST:
        this->printCastExp(node);
        break;
    case ExpKind::GETINT:
        this->printGetintExp();
        break;
    case ExpKind::GETCH:
        this->printGetchExp();
        break;
    case ExpKind::GETFLOAT:
        this->printGetfloatExp();
        break;
    case ExpKind::GETARRAY:
        this->printGetarrayExp(node);
        break;
    case ExpKind::GETFARRAY:
        this->printGetfarrayExp(node);
        break;
    case ExpKind::ALLOCA:
        this->printAllocaExp(node);
        break;
    default:
        break;
    }
}

void Printer::printBinOpExp(Exp *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<BinOp>\n";
    this->indent++;
    this->os << std::string(this->indent, '\t') << "<type>" << getExpTypeStr(node->type) << "</type>\n";
    switch (node->u.binop.op) {
    case BinOp::ADD:
        this->os << std::string(this->indent, '\t') << "<Op>ADD</Op>\n";
        break;
    case BinOp::SUB:
        this->os << std::string(this->indent, '\t') << "<Op>SUB</Op>\n";
        break;
    case BinOp::MUL:
        this->os << std::string(this->indent, '\t') << "<Op>MUL</Op>\n";
        break;
    case BinOp::DIV:
        this->os << std::string(this->indent, '\t') << "<Op>DIV</Op>\n";
        break;
    case BinOp::MOD:
        this->os << std::string(this->indent, '\t') << "<Op>MOD</Op>\n";
        break;
    default:
        break;
    }
    this->printExp(node->u.binop.left);
    this->printExp(node->u.binop.right);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</BinOp>\n";
}

void Printer::printMemExp(Exp *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Mem>\n";
    this->indent++;
    this->os << std::string(this->indent, '\t') << "<type>" << getExpTypeStr(node->type) << "</type>\n";
    this->printExp(node->u.mem);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Mem>\n";
}

void Printer::printTempExp(Exp *node) {
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Printing TempExp" << std::endl;
#endif
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Temp>\n";
    this->indent++;
    this->os << std::string(this->indent, '\t') << "<type>" << getExpTypeStr(node->type) << "</type>\n";
    this->printTemp(node->u.temp);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Temp>\n";
}

void Printer::printEseqExp(Exp *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Eseq>\n";
    this->indent++;
    this->os << std::string(this->indent, '\t') << "<type>" << getExpTypeStr(node->type) << "</type>\n";
    this->printStmt(node->u.eseq.stmt);
    this->printExp(node->u.eseq.exp);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Eseq>\n";
}

void Printer::printNameExp(Exp *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Name>" << node->u.name->name << "</Name>\n";
}

void Printer::printConstIExp(Exp *node) {
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Printing ConstIExp" << std::endl;
#endif
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<ConstI>" << node->u.consti << "</ConstI>\n";
#ifdef IRP_PRINTER_DEBUG
    std::cerr << "Finished printing ConstIExp" << std::endl;
#endif
}

void Printer::printConstFExp(Exp *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<ConstF>" << node->u.constf << "</ConstF>\n";
}

void Printer::printCallExp(Exp *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Call>\n";
    this->indent++;
    this->os << std::string(this->indent, '\t') << "<type>" << getExpTypeStr(node->type) << "</type>\n";
    this->os << std::string(this->indent, '\t') << "<name>" << *node->u.call.name << "</name>\n";
    this->printExpList(node->u.call.args);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Call>\n";
}

void Printer::printCastExp(Exp *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Cast>\n";
    this->indent++;
    this->os << std::string(this->indent, '\t') << "<type>" << getExpTypeStr(node->type) << "</type>\n";
    this->printExp(node->u.cast);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Cast>\n";
}

void Printer::printGetintExp() { this->os << std::string(this->indent, '\t') << "<Getint/>\n"; }

void Printer::printGetchExp() { this->os << std::string(this->indent, '\t') << "<Getch/>\n"; }

void Printer::printGetfloatExp() { this->os << std::string(this->indent, '\t') << "<Getfloat/>\n"; }

void Printer::printGetarrayExp(Exp *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Getarray>\n";
    this->indent++;
    this->os << std::string(this->indent, '\t') << "<type>" << getExpTypeStr(node->type) << "</type>\n";
    this->printExp(node->u.getarray);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Getarray>\n";
}

void Printer::printGetfarrayExp(Exp *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Getfarray>\n";
    this->indent++;
    this->os << std::string(this->indent, '\t') << "<type>" << getExpTypeStr(node->type) << "</type>\n";
    this->printExp(node->u.getarray);
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Getfarray>\n";
}

void Printer::printAllocaExp(Exp *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<Alloca>\n";
    this->indent++;
    this->os << std::string(this->indent, '\t') << "<type>" << getExpTypeStr(node->type) << "</type>\n";
    this->os << std::string(this->indent, '\t') << "<size>" << node->u.alloca.size << "</size>\n";
    this->os << std::string(this->indent, '\t') << "<offset>" << node->u.alloca.frameOffset << "</offset>\n";
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</Alloca>\n";
}

void Printer::printExpList(ExpList *node) {
    if (!node) {
        return;
    }
    this->os << std::string(this->indent, '\t') << "<args>\n";
    this->indent++;
    for (auto &exp : *node) {
        this->printExp(exp);
    }
    this->indent--;
    this->os << std::string(this->indent, '\t') << "</args>\n";
}
} // namespace irp