#include "stmPrinter.hpp"
#include <iostream>
#include <ostream>

#define STM_PRINTER_DEBUG
#undef STM_PRINTER_DEBUG

namespace irp {
static void printStmt(std::ostream &os, Stmt *stmt, int indent);
static void printExp(std::ostream &os, Exp *exp);
static void printExpList(std::ostream &os, ExpList *expList);
static void printTemp(std::ostream &os, temp::Temp *temp);

void printStmtList(std::ostream &os, StmtList *stmtList, int indent) {
#ifdef STM_PRINTER_DEBUG
    std::cerr << "printStmtList" << std::endl;
#endif
    auto sl = stmtList;
    while (sl) {
        printStmt(os, sl->head, indent + 1);
        os << std::endl;
        sl = sl->tail;
    }
}

static void printStmt(std::ostream &os, irp::Stmt *stmt, int indent) {
    if (!stmt) {
        return;
    }
#ifdef STM_PRINTER_DEBUG
    std::cerr << "printStmt" << std::endl;
#endif
    switch (stmt->kind) {
    case StmtKind::SEQ: {
        os << std::string(indent, '\t');
        os << "(";
        printStmt(os, stmt->u.seq.left, indent + 1);
        os << ", ";
        printStmt(os, stmt->u.seq.right, indent + 1);
        os << ")";
        break;
    }
    case StmtKind::LABEL: {
        os << stmt->u.label->name;
        break;
    }
    case StmtKind::MOVE: {
        os << std::string(indent, '\t');
        os << "Move ";
        printExp(os, stmt->u.move.dst);
        os << ", ";
        printExp(os, stmt->u.move.src);
        break;
    }
    case StmtKind::EXP: {
        os << std::string(indent, '\t');
        printExp(os, stmt->u.exp);
        break;
    }
    case StmtKind::RETURN: {
        os << std::string(indent, '\t');
        os << "Return ";
        printExp(os, stmt->u.exp);
        break;
    }
    case StmtKind::JUMP: {
        os << std::string(indent, '\t');
        os << "Jump ";
        os << stmt->u.jump->name;
        break;
    }
    case StmtKind::CJUMP: {
        os << std::string(indent, '\t');
        os << "CJump ";
        switch (stmt->u.cjump.op) {
        case RelOp::EQ:
            os << "==";
            break;
        case RelOp::NE:
            os << "!=";
            break;
        case RelOp::LT:
            os << "<";
            break;
        case RelOp::GT:
            os << ">";
            break;
        case RelOp::LE:
            os << "<=";
            break;
        case RelOp::GE:
            os << ">=";
            break;
        }
        os << " ";
        printExp(os, stmt->u.cjump.left);
        os << ", ";
        printExp(os, stmt->u.cjump.right);
        os << ", ";
        os << stmt->u.cjump.t->name;
        os << ", ";
        os << stmt->u.cjump.f->name;
        break;
    }
    case StmtKind::VOID_CALL: {
        os << std::string(indent, '\t');
        os << *stmt->u.void_call.name;
        os << "(";
        printExpList(os, stmt->u.void_call.args);
        os << ")";
        break;
    }
    case StmtKind::PUTINT: {
        os << std::string(indent, '\t');
        os << "Putint(";
        printExp(os, stmt->u.exp);
        os << ")";
        break;
    }
    case StmtKind::PUTCH: {
        os << std::string(indent, '\t');
        os << "Putch(";
        printExp(os, stmt->u.exp);
        os << ")";
        break;
    }
    case StmtKind::PUTFLOAT: {
        os << std::string(indent, '\t');
        os << "Putfloat(";
        printExp(os, stmt->u.exp);
        os << ")";
        break;
    }
    case StmtKind::PUTARRAY: {
        os << std::string(indent, '\t');
        os << "Putarray(";
        printExp(os, stmt->u.putarray.outlen);
        os << ", ";
        printExp(os, stmt->u.putarray.array);
        os << ")";
        break;
    }
    case StmtKind::PUTFARRAY: {
        os << std::string(indent, '\t');
        os << "Putfarray(";
        printExp(os, stmt->u.putarray.outlen);
        os << ", ";
        printExp(os, stmt->u.putarray.array);
        os << ")";
        break;
    }
    case StmtKind::PUTF: {
        os << std::string(indent, '\t');
        os << "Putf ";
        os << "\"" << *stmt->u.putf.fmt << "\"";
        os << "(";
        printExpList(os, stmt->u.putf.args);
        os << ")";
        break;
    }
    case StmtKind::STARTTIME: {
        os << std::string(indent, '\t');
        os << "Starttime";
        break;
    }
    case StmtKind::STOPTIME: {
        os << std::string(indent, '\t');
        os << "Stoptime";
        break;
    }
    case StmtKind::MEMSET: {
        os << std::string(indent, '\t');
        os << "Memset ";
        printTemp(os, stmt->u.memset.array);
        os << ", ";
        os << stmt->u.memset.offset;
        os << ", ";
        os << stmt->u.memset.size;
        break;
    }
    }
}

static void printExp(std::ostream &os, Exp *exp) {
    if (!exp) {
        return;
    }
    switch (exp->kind) {
    case ExpKind::BINOP: {
        os << "Binop(";
        switch (exp->u.binop.op) {
        case BinOp::ADD:
            os << "ADD";
            break;
        case BinOp::SUB:
            os << "SUB";
            break;
        case BinOp::MUL:
            os << "MUL";
            break;
        case BinOp::DIV:
            os << "DIV";
            break;
        case BinOp::MOD:
            os << "MOD";
            break;
        }
        os << ", ";
        printExp(os, exp->u.binop.left);
        os << ", ";
        printExp(os, exp->u.binop.right);
        os << ")";
        break;
    }
    case ExpKind::MEM: {
        os << "Mem(";
        printExp(os, exp->u.mem);
        os << ")";
        break;
    }
    case ExpKind::TEMP: {
        printTemp(os, exp->u.temp);
        break;
    }
    case ExpKind::ESEQ: {
        os << "Eseq(";
        printStmt(os, exp->u.eseq.stmt, 0);
        os << ",\n";
        printExp(os, exp->u.eseq.exp);
        os << ")";
        break;
    }
    case ExpKind::NAME: {
        os << exp->u.name->name;
        break;
    }
    case ExpKind::CONSTI: {
        os << exp->u.consti;
        break;
    }
    case ExpKind::CONSTF: {
        os << exp->u.constf;
        break;
    }
    case ExpKind::CALL: {
        os << *exp->u.call.name;
        os << "(";
        printExpList(os, exp->u.call.args);
        os << ")";
        break;
    }
    case ExpKind::CAST: {
        os << "Cast(";
        printExp(os, exp->u.cast);
        os << ")";
        break;
    }
    case ExpKind::GETINT: {
        os << "Getint(";
        printExp(os, exp->u.getarray);
        os << ")";
        break;
    }
    case ExpKind::GETFLOAT: {
        os << "Getfloat(";
        printExp(os, exp->u.getarray);
        os << ")";
        break;
    }
    case ExpKind::GETCH: {
        os << "Getch(";
        printExp(os, exp->u.getarray);
        os << ")";
        break;
    }
    case ExpKind::GETARRAY: {
        os << "Getarray(";
        printExp(os, exp->u.getarray);
        os << ")";
        break;
    }
    case ExpKind::GETFARRAY: {
        os << "Getfarray(";
        printExp(os, exp->u.getarray);
        os << ")";
        break;
    }
    case ExpKind::ALLOCA: {
        os << "Alloca(";
        os << exp->u.alloca.size;
        os << ")";
        break;
    }
    }
    os << ":";
    switch (exp->type) {
    case ExpType::INT:
        os << "INT";
        break;
    case ExpType::FLOAT:
        os << "FLOAT";
        break;
    }
}

static void printExpList(std::ostream &os, ExpList *expList) {
    if (!expList) {
        return;
    }
    for (int i = 0; i < expList->size(); ++i) {
        printExp(os, expList->at(i));
        if (i != expList->size() - 1) {
            os << ", ";
        }
    }
}

static void printTemp(std::ostream &os, temp::Temp *temp) { os << "t" << temp->num; }
} // namespace irp